__author__ = "Ruben Svensson"
__copyright__ = """

    Copyright 2017-2019 Modern Ancient Instruments Networked AB, dba Elk

    elkpy is free software: you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation, either version 3
    of the License, or (at your option) any later version.

    elkpy is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with elkpy.  If
    not, see <http://www.gnu.org/licenses/>.
"""
__license__ = "GPL-3.0"

import grpc

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from elkpy.sushicontroller import SushiController

from . import sushierrors
from . import grpc_gen
from . import sushi_info_types as info_types

####################################
# Sushi transport controller class #
####################################


class TransportController(object):
    """
    A class to control the transport in sushi via gRPC. It controls transport related information in sushi
    like samplerate, playback mode, sync mode, time signature and tempo.

    Attributes:
        _stub (TransportControllerStub): Connection stubs to the gRPC transport interface implemented in sushi.
    """

    def __init__(
        self,
        parent: 'SushiController',
        address="localhost:51051",
        sushi_proto_def="/usr/share/sushi/sushi_rpc.proto",
        sushi_proto = None, sushi_grpc = None
    ):
        """
        The constructor for the TransportController class setting up the gRPC connection with sushi.

        Parameters:
            address (str): 'ip-addres:port' The ip-addres and port at which to connect to sushi.
            sushi_proto_def (str): path to .proto file with SUSHI's gRPC services definition
        """
        try:
            channel = grpc.insecure_channel(address)
        except AttributeError as e:
            raise TypeError(
                "Parameter address = {}. Should be a string containing the ip-address and port of sushi ('ip-address:port')".format(
                    address
                )
            ) from e

        if sushi_proto and sushi_grpc:
            self._sushi_proto = sushi_proto
            self._sushi_grpc = sushi_grpc
        else:
            self._sushi_proto, self._sushi_grpc = grpc_gen.modules_from_proto(sushi_proto_def)
        
        self._stub = self._sushi_grpc.TransportControllerStub(channel)
        self._parent = parent 

    def get_samplerate(self) -> float:
        """
        Get the current samplerate.

        Returns:
            float: Current samplerate.
        """
        try:
            response = self._stub.GetSamplerate(self._sushi_proto.GenericVoidValue())
            return response.value

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def get_playing_mode(self) -> int:
        """
        Get the current playing mode.

        Returns:
            int: Current playing mode.
                1 = Stopped,
                2 = Playing,
                3 = Recording (not implemented)
        """
        try:
            response = self._stub.GetPlayingMode(self._sushi_proto.GenericVoidValue())
            return info_types.PlayingMode(response.mode)

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def set_playing_mode(self, playing_mode: info_types.PlayingMode) -> None:
        """
        Set the playing mode.

        Parameters:
            playing_mode (PlayingMode): The playing mode to set.
                                1 = Stopped,
                                2 = Playing,
                                3 = Recording (not implemented)
        """

        if info_types.PlayingMode(playing_mode) in info_types.PlayingMode:
            try:
                cr = self._stub.SetPlayingMode(self._sushi_proto.PlayingMode(mode=int(playing_mode)))
                return self._parent.validate_command_response(cr, context_info=f"Setting playing mode to {self._sushi_proto.PlayingMode(mode=playing_mode)}")

            except grpc.RpcError as e:
                sushierrors.grpc_error_handling(
                    e, " With playing mode: {}".format(playing_mode)
                )

    def get_sync_mode(self) -> info_types.SyncMode:
        """
        Get the current sync mode.

        Returns:
            int: Current sync mode.
                1 = Internal,
                2 = MIDI,
                3 = Link
        """
        try:
            response = self._stub.GetSyncMode(self._sushi_proto.GenericVoidValue())
            return info_types.SyncMode(response.mode)

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def set_sync_mode(self, sync_mode: info_types.SyncMode) -> None:
        """
        Set the sync mode.

        Parameters:
            sync_mode (SyncMode): The sync mode to set.
                            1 = Internal,
                            2 = MIDI,
                            3 = Link
        """
        if info_types.SyncMode(sync_mode) in info_types.SyncMode:
            try:
                cr = self._stub.SetSyncMode(self._sushi_proto.SyncMode(mode=int(sync_mode)))
                return self._parent.validate_command_response(cr, context_info=f"Setting sync mode to {self._sushi_proto.SyncMode(mode=int(sync_mode))}")

            except grpc.RpcError as e:
                sushierrors.grpc_error_handling(
                    e, " With sync mode: {}".format(sync_mode)
                )

    def get_tempo(self) -> float:
        """
        Get the current tempo.

        Returns:
            float: Current tempo in BPM(Beats Per Minute).
        """
        try:
            response = self._stub.GetTempo(self._sushi_proto.GenericVoidValue())
            return response.value
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def set_tempo(self, tempo: float) -> None:
        """
        Set the tempo.

        Parameters:
            tempo (float): The tempo in BPM(Beats Per Minute).
        """
        try:
            cr = self._stub.SetTempo(self._sushi_proto.GenericFloatValue(value=tempo))
            return self._parent.validate_command_response(cr, context_info=f"Setting tempo to {tempo}")

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, " With tempo: {}".format(tempo))

    def get_time_signature(self) -> tuple[int, int]:
        """
        Get the current time signature.

        Returns:
            int: The nominator of the time signature.
            int: The denominator of the time signature.
        """
        try:
            response = self._stub.GetTimeSignature(self._sushi_proto.GenericVoidValue())
            return response.numerator, response.denominator

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def set_time_signature(self, numerator: int, denominator: int) -> None:
        """
        Set the time signature

        Parameters:
            numerator (int): The numerator of the time signature.
            denominator (int): The denominator of the time signature. Should be either 4 or 8.
        """
        try:
            cr = self._stub.SetTimeSignature(self._sushi_proto.TimeSignature(numerator=numerator, denominator=denominator))
            return self._parent.validate_command_response(cr, context_info=f"Setting time signature to {self._sushi_proto.TimeSignature(numerator=numerator, denominator=denominator)}")

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(
                e, " With numerator: {}, denominator: {}".format(numerator, denominator)
            )



__author__ = "Ruben Svensson"
__copyright__ = """

    Copyright 2017-2019 Modern Ancient Instruments Networked AB, dba Elk

    elkpy is free software: you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation, either version 3
    of the License, or (at your option) any later version.

    elkpy is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with elkpy.  If
    not, see <http://www.gnu.org/licenses/>.
"""
__license__ = "GPL-3.0"


############################
# Error handling functions #
############################

from typing import NoReturn


class SushiUnkownError(Exception):
    pass


class SushiUnsupportedOperationError(Exception):
    pass


class SushiNotFoundError(Exception):
    pass


class SushiOutOfRangeError(Exception):
    pass


class SushiInvalidArgumentError(Exception):
    pass


class SushiInternalError(Exception):
    pass


class SushiUnavailableError(Exception):
    pass


class SushiUnimplementedError(Exception):
    pass


def grpc_error_handling(e, context_info="") -> NoReturn:
    if e.code().name == "UNKNOWN":
        raise SushiUnkownError(e.details(), context_info) from e
    elif e.code().name == "UNAVAILABLE":
        raise SushiUnavailableError(e.details(), context_info) from e
    elif e.code().name == "UNIMPLEMENTED":
        raise SushiUnimplementedError(e.details(), context_info) from e
    else:
        if context_info != "":
            print(context_info)
        raise e



__author__ = "Ruben Svensson"
__copyright__ = """

    Copyright 2017-2019 Modern Ancient Instruments Networked AB, dba Elk

    elkpy is free software: you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation, either version 3
    of the License, or (at your option) any later version.

    elkpy is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with elkpy.  If
    not, see <http://www.gnu.org/licenses/>.
"""
__license__ = "GPL-3.0"

import __main__
import os
import sys
import importlib
import grpc_tools.protoc as gprotoc
from types import ModuleType
from typing import Tuple 


def modules_from_proto(proto_filename: str) -> Tuple[ModuleType, ModuleType]:
    """
    Run protoc compiler and get generated modules.

    Parameters:
        proto_filename : path to .proto file with service definition

    Returns:
        (protobuf_module, grpc_module)
    """
    full_path = os.path.abspath(proto_filename)
    [inc_path, rel_proto_filename] = os.path.split(full_path)
    out_dir = "."
    if hasattr(__main__, "__file__"):
        file_path = os.path.dirname(os.path.abspath(__main__.__file__))
        if sys.path[0] == file_path:
            out_dir = file_path

    protoc_args = [ 'dummy',
                    '-I%s' % inc_path,
                    f'--python_out={out_dir}',
                    f'--grpc_python_out={out_dir}',
                    rel_proto_filename ]
    gprotoc.main(protoc_args)
    proto_base_name = os.path.splitext(rel_proto_filename)[0]
    proto_module_name = '%s_pb2' % proto_base_name
    proto_module = importlib.import_module(proto_module_name)
    grpc_module_name = '%s_pb2_grpc' % proto_base_name
    grpc_module = importlib.import_module(grpc_module_name)
    return (proto_module, grpc_module)



__author__ = "Ruben Svensson"
__copyright__ = """

    Copyright 2017-2019 Modern Ancient Instruments Networked AB, dba Elk

    elkpy is free software: you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation, either version 3
    of the License, or (at your option) any later version.

    elkpy is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with elkpy.  If
    not, see <http://www.gnu.org/licenses/>.
"""
__license__ = "GPL-3.0"

from .sushicontroller import SushiController
from typing import List

class SushiProcessor:
    """
    A Class to provide control of a specific processor in sushi.

    Attributes:
        _name (str): The name of the processor.
        _controller (SushiController): The sushi controller to use.
        _track_id (int): The id of the track the processor is assigned to.
        _id (int): The id corresponding to the name of the processor.
        _parameters (dict): A mapping from parameter name to parameter id.
        _programs (dict): A mapping from program name to program id.
    """
    def __init__(self, processor_name: str, controller: SushiController):
        """
        Constructor for the sushi processor. Takes the name and the controller used to control sushi.

        Parameters:
            processor_name (str): The name of the processor to control.
            controller (SushiController): The controller to use for controlling the processor.
        """
        self._name = processor_name
        self._controller = controller
        self._track_id = -1
        self._id = self._controller.audio_graph.get_processor_id(self._name)
        self._parameters = {}
        self._programs = {}


        # TODO: Use try block when error handling is approved
        for parameter in controller.parameters.get_processor_parameters(self._id):
            self._parameters[parameter.name] = parameter.id

        if (self._controller.audio_graph.get_processor_info(self._id).program_count > 0):
            # TODO: Use try block when error handling is approved
            for program in controller.programs.get_processor_programs(self._id):
                self._programs[program.name] = program.id

    #####################
    # Parameter Control #
    #####################

    def set_parameter_value(self, parameter_name: str, value: float) -> None:
        """
        Set the value of parameter by name.

        Parameters:
            parameter_name (str): The name of the parameter to set the value of.
            value (float): The value to set the parameter to.
        """
        self._controller.parameters.set_parameter_value(self._id, self._parameters[parameter_name], value)

    def get_parameter_value(self, parameter_name: str) -> float:
        """
        Get the value of a parameter by name.

        Parameters:
            parameter_name (str): The name of the parameter to get the value from.

        Returns:
            float: The current value of the parameter.
        """
        return self._controller.parameters.get_parameter_value(self._id, self._parameters[parameter_name])

    def get_parameters(self) -> List[str]:
        """
        Get a list of the names of the parameters available to the processor.

        Returns:
            List[str]: List of parameter names.
        """
        return list(self._parameters)

    def get_parameter_values(self) -> dict:
        """
        Get the current value of the parameters with their name as the key.

        Returns:
            dict: Dictionary with key as parameter names and value as the current parameter value.
        """
        parameter_values = {}
        for param in self._parameters:
            parameter_values[param] = self._controller.parameters.get_parameter_value(self._id, self._parameters[param])

        return parameter_values

    def get_bypass_state(self) -> bool:
        """
        Get the bypass state of the processor.

        Returns:
            bool: The bypass state of the processor.
        """
        return self._controller.audio_graph.get_processor_bypass_state(self._id)

    def set_bypass_state(self, bypass_state: bool) -> None:
        """
        Set the bypass state of the processor.

        Parameters:
            bypass_state (bool): The bypass state to set the processor to.
        """
        self._controller.audio_graph.set_processor_bypass_state(self._id, bypass_state)

    ###################
    # Program control #
    ###################

    def set_program(self, program_name: str) -> None:
        """
        Set the current program of the processor with the program name

        Parameters:
            program_name (str): The name of the program to set the processor to.
        """
        try:
            self._controller.programs.set_processor_program(self._id, self._programs[program_name])
        except KeyError:
            pass

    def set_program_next(self):
        """
        Set the processor to the next program or loopback to the beginning if at the end of the program list
        """
        number_of_programs = len(self._programs)
        current_program_index = self._controller.programs.get_processor_current_program(self._id)

        if current_program_index == number_of_programs-1:
            new_program_index = 0
        else:
            new_program_index = current_program_index + 1

        self._controller.programs.set_processor_program(self._id, new_program_index)

    def set_program_previous(self):
        """
        Set the processor to the previous program or loopback to the end if at the start of the program list
        """
        number_of_programs = len(self._programs)
        current_program_index = self._controller.programs.get_processor_current_program(self._id)

        if current_program_index == 0:
            new_program_index = number_of_programs-1
        else:
            new_program_index = current_program_index - 1

        self._controller.programs.set_processor_program(self._id, new_program_index)

    def get_program(self) -> str:
        """
        Get the name of the current program.

            Returns (str): The name of the current program.
        """
        return self._controller.programs.get_processor_current_program_name(self._id)

    def get_programs(self) -> List[str]:
        """
        Get a list of the names of the available programs.

            Returns:
                List[str]: The names of the available programs.
        """
        return list(self._programs)

    ####################
    # Keyboard control #
    ####################

    def send_note_on(self, channel: int, note: int, velocity: float):
        """
        Send a note on message to the track containing the processor.

        Parameters:
            channel (int): The channel to send the message on.
            note (int): The midi note value to send.
            velocity (float): The velocity of the note as a float between 0-1.
        """
        self._controller.keyboard.send_note_on(self._track_id, channel, note, velocity)

    def send_note_off(self, channel: int, note: int, velocity: float):
        """
        Send a note off message to the track containing the processor.

        Parameters:
            channel (int): The channel to send the message on.
            note (int): The midi note value to send.
            velocity (float): The velocity of the note as a float between 0-1.
        """
        self._controller.keyboard.send_note_off(self._track_id, channel, note, velocity)

    def send_note_aftertouch(self, channel: int, note: int, value: float):
        """
        Send a note aftertouch message to the track containing the processor.

        Parameters:
            channel (int): The channel to send the message on.
            note (int): The midi note value to send.
            value (float): The aftertouch value of the note as a float between 0-1.
        """
        self._controller.keyboard.send_note_aftertouch(self._track_id, channel, note, value)

    def send_aftertouch(self, channel: int, value: float):
        """
        Send a aftertouch message to the track containing the processor.

        Parameters:
            channel (int): The channel to send the message on.
            value (float): The aftertouch value of the note as a float between 0-1.
        """
        self._controller.keyboard.send_aftertouch(self._track_id, channel, value)

    def send_pitch_bend(self, channel: int, value: float):
        """
        Send a pitch bend message to the track containing the processor.

        Parameters:
            channel (int): The channel to send the message on.
            value (float): The pitch bend value of the note as a float between 0-1.
        """
        self._controller.keyboard.send_pitch_bend(self._track_id, channel, value)

    def send_modulation(self, channel: int, value: float):
        """
        Send a modulation message to the track containing the processor.

        Parameters:
            channel (int): The channel to send the message on.
            value (float): The modulation value of the note as a float between 0-1.
        """
        self._controller.keyboard.send_pitch_bend(self._track_id, channel, value)



__author__ = "Maxime Gendebien"
__copyright__ = """

    Copyright 2017-2019 Modern Ancient Instruments Networked AB, dba Elk

    elkpy is free software: you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation, either version 3
    of the License, or (at your option) any later version.

    elkpy is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with elkpy.  If
    not, see <http://www.gnu.org/licenses/>.
"""
__license__ = "GPL-3.0"

import grpc
from . import sushierrors
from . import sushi_info_types as info_types
from . import grpc_gen

###############################
#   System Controller class   #
###############################


class SystemController(object):
    """Class to manage system-level settings in Sushi"""

    def __init__(
        self,
        address: str = "localhost:51051",
        sushi_proto_def: str = "/usr/share/sushi/sushi_rpc.proto",
    ) -> None:
        """
        Args:
            address: IP address to Sushi in the uri form : 'ip-addr:port'
            sushi_proto_def: path to the .proto file with SUSHI gRPC services definitions
        """
        try:
            channel = grpc.insecure_channel(address)
        except AttributeError as e:
            raise TypeError(
                f"Parameter address = {address}. Should be a string containing the ip-address and port "
                f"to Sushi"
            ) from e

        self._sushi_proto, self._sushi_grpc = grpc_gen.modules_from_proto(
            sushi_proto_def
        )
        self._stub = self._sushi_grpc.SystemControllerStub(channel)

    def get_sushi_version(self) -> str:
        try:
            response = self._stub.GetSushiVersion(self._sushi_proto.GenericVoidValue())
            return response.value
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def get_sushi_api_version(self) -> str:
        try:
            response = self._stub.GetSushiApiVersion(self._sushi_proto.GenericVoidValue())
            return response.value
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def get_build_info(self) -> info_types.SushiBuildInfo:
        try:
            response = self._stub.GetBuildInfo(self._sushi_proto.GenericVoidValue())
            return info_types.SushiBuildInfo(response)
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def get_input_audio_channel_count(self) -> int:
        try:
            response = self._stub.GetInputAudioChannelCount(
                self._sushi_proto.GenericVoidValue()
            )
            return response.value
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def get_output_audio_channel_count(self) -> int:
        try:
            response = self._stub.GetOutputAudioChannelCount(
                self._sushi_proto.GenericVoidValue()
            )
            return response.value
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)



__author__ = "Maxime Gendebien"
__copyright__ = """

    Copyright 2017-2019 Modern Ancient Instruments Networked AB, dba Elk

    elkpy is free software: you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation, either version 3
    of the License, or (at your option) any later version.

    elkpy is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with elkpy.  If
    not, see <http://www.gnu.org/licenses/>.
"""
__license__ = "GPL-3.0"

from typing import TYPE_CHECKING, List
if TYPE_CHECKING:
    from elkpy.sushicontroller import SushiController

from .events import SushiCommandResponse
import grpc

from . import sushierrors
from . import sushi_info_types as info_types
from . import grpc_gen

############################################
#   Sushi MIDI Controller class   #
############################################


class MidiController(object):
    """
    A class to manage MIDI connections in Sushi via gRPC.
    It can create and remove keyboard, CC and PC connections from an external midi device to tracks,
    processors, and parameters in sushi.
    """
    def __init__(self,
                 parent: 'SushiController',
                 address: str='localhost:51051',
                 sushi_proto_def: str='/usr/share/sushi/sushi_rpc.proto',
                 sushi_proto = None, sushi_grpc = None) -> None:
        """
        The constructor for the MidiController class setting up the gRPC connection with sushi.

        Parameters:
            address (str): IP address to Sushi in the uri form : 'ip-addr:port'
            sushi_proto_def (str): path to the .proto file with SUSHI gRPC services definitions
        """
        try:
            channel = grpc.insecure_channel(address)
        except AttributeError as e:
            raise TypeError(f"Parameter address = {address}. Should be a string containing the ip-address and port "
                            f"to Sushi") from e

        if sushi_proto and sushi_grpc:
            self._sushi_proto = sushi_proto
            self._sushi_grpc = sushi_grpc
        else:
            self._sushi_proto, self._sushi_grpc = grpc_gen.modules_from_proto(sushi_proto_def)
        self._stub = self._sushi_grpc.MidiControllerStub(channel)
        self._parent = parent

    def get_input_ports(self) -> int:
        """
        Gets the number of MIDI input ports.

        Returns:
            int: The number of MIDI input ports enabled in Sushi
        """
        try:
            response = self._stub.GetInputPorts(self._sushi_proto.GenericVoidValue())
            return response.value
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def get_output_ports(self) -> int:
        """
        Gets the number of MIDI output ports.

        Returns:
            int: The number of MIDI output ports enabled in Sushi
        """
        try:
            response = self._stub.GetOutputPorts(self._sushi_proto.GenericVoidValue())
            return response.value
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def get_all_kbd_input_connections(self) -> List[info_types.MidiKbdConnection]:
        """
        Gets a list of all MIDI Keyboard input Connections in Sushi

        Returns:
            List[info_types.MidiKbdConnection]: List of MidiKbdConnection objects
        """
        try:
            response = self._stub.GetAllKbdInputConnections(self._sushi_proto.GenericVoidValue())
            return [info_types.MidiKbdConnection(c) for c in response.connections]
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def get_all_kbd_output_connections(self) -> List[info_types.MidiKbdConnection]:
        """
        Gets a list of all MIDI Keyboard output Connections in Sushi

        Returns:
            List[info_types.MidiKbdConnection]: List of MidiKbdConnection objects
        """
        try:
            response = self._stub.GetAllKbdOutputConnections(self._sushi_proto.GenericVoidValue())
            return [info_types.MidiKbdConnection(c) for c in response.connections]
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def get_all_cc_input_connections(self) -> List[info_types.MidiCCConnection]:
        """
        Gets a list of all MIDI Control Change input connections in Sushi

        Returns:
            List[info_types.MidiCCConnection]: List of MidiCCConnection objects
        """
        try:
            response = self._stub.GetAllCCInputConnections(self._sushi_proto.GenericVoidValue())
            return [info_types.MidiCCConnection(c) for c in response.connections]
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def get_all_pc_input_connections(self) -> List[info_types.MidiPCConnection]:
        """
        Gets a list of all MIDI Program Change input connections in Sushi

        Returns:
            List[info_types.MidiPCConnection]: List of MidiPCConnection objects
        """
        try:
            response = self._stub.GetAllPCInputConnections(self._sushi_proto.GenericVoidValue())
            return [info_types.MidiPCConnection(c) for c in response.connections]
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def get_cc_input_connections_for_processor(self, processor_id: int) -> List[info_types.MidiCCConnection]:
        """
        Gets a list of all MIDI Control Change connection for a processor.

        Parameters:
            processor_id (int): The id of the processor to get the input connections from

        Returns:
            List[info_types.MidiCCConnection]: List of MidiCCConnection objects
        """
        try:
            cr = self._stub.GetCCInputConnectionsForProcessor(self._sushi_proto.ProcessorIdentifier(id=processor_id))
            response = self._parent.validate_command_response(cr, f"Get CC input connections for processor {processor_id}")
            return [info_types.MidiCCConnection(c) for c in response.connections]
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, f"With processor_id: {processor_id}")

    def get_pc_input_connections_for_processor(self, processor_id: int) -> List[info_types.MidiPCConnection]:
        """
        Gets a list of all MIDI Program Change connection for a processor.

        Parameters:
            processor_id (int): The id of the processor to get the input connections from

        Returns:
            List[info_types.MidiPCConnection]: List of MidiPCConnection objects
        """
        try:
            cr = self._stub.GetPCInputConnectionsForProcessor(
                self._sushi_proto.ProcessorIdentifier(id=processor_id))
            response = self._parent.validate_command_response(cr, f"Get PC input connections for processor {processor_id}")
            return [info_types.MidiPCConnection(c) for c in response.connections]
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, f"With processor_id: {processor_id}")

    def get_midi_clock_output_enabled(self, port: int) -> bool:
        """
        Gets whether MIDI clock output is enabled for a midi port.

        Parameters:
            port (int): The id of the processor to query

        Returns:
            bool: true if midi clock is enabled for that midi output port
        """
        try:
            response = self._stub.GetMidiClockOutputEnabled(self._sushi_proto.GenericIntValue(value=port))
            return response.value
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, f"With port: {port}")

    def set_midi_clock_output_enabled(self, port: int, enabled: bool) -> SushiCommandResponse:
        """
        Enable or disable MIDI clock output for a midi port.

        Parameters:
            port (int): The id of the output port to toggle
            enabled (bool): True to turn on clock output, False to turn off

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.SetMidiClockOutputEnabled(
                self._sushi_proto.MidiClockSetRequest(port=port, enabled=enabled))
            return self._parent.validate_command_response(cr, f"Set MIDI clock output to {enabled}")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, f"With port: {port}, enabled: {enabled}")

    def connect_kbd_input_to_track(self, track: int, channel: int, port: int, raw_midi: bool) -> SushiCommandResponse:
        """
        Connects MIDI Keyboard messages to a track

        Parameters:
            track (int): The id of the track to connect to
            channel (int): The id of the channel to connect to
            port (int): The id of the port to connect to
            raw_midi (bool): Enable raw MIDI

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.ConnectKbdInputToTrack(self._sushi_proto.MidiKbdConnection(track=self._sushi_proto.TrackIdentifier(id=track),
                                                                                  channel=self._sushi_proto.MidiChannel(channel=channel),
                                                                                  port=port,
                                                                                  raw_midi=raw_midi))
            return self._parent.validate_command_response(cr, "Connect Keyboard input to track")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, f"With MidiKbdConnection: track: {track}, channel: {channel}, "
                                               f"port: {port}, raw_midi: {raw_midi}")

    def connect_kbd_output_from_track(self, track: int, channel: int, port: int, raw_midi: bool) -> SushiCommandResponse:
        """
        Connects MIDI Keyboard messages from a track to a MIDI output port

        Parameters:
            track (int): The id of the track to connect to
            channel (int): The id of the channel to connect to
            port (int): The id of the port to connect to
            raw_midi (bool): Enable raw MIDI

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.ConnectKbdOutputFromTrack(self._sushi_proto.MidiKbdConnection(track=self._sushi_proto.TrackIdentifier(id=track),
                                                                                     channel=self._sushi_proto.MidiChannel(channel=channel),
                                                                                     port=port,
                                                                                     raw_midi=raw_midi))
            return self._parent.validate_command_response(cr, "Connect Keyboard output to track")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, f"With MidiKbdConnection: track: {track}, channel: {channel}, "
                                               f"port: {port}, raw_midi: {raw_midi}")

    def connect_cc_to_parameter(self, processor_id: int, parameter_id: int, channel: int, port: int, cc_number: int,
                                min_range: float, max_range: float, relative_mode: bool) -> SushiCommandResponse:
        """
        Connects MIDI Control Change messages to a parameter

        Parameters:
            processor_id (int): The id of processor that the parameter belongs to
            parameter_id (int): The id of the parameter to connect to
            channel (int): The MIDI channel to use for the connection
            port (int): The id of the MIDI port to use for the connection
            cc_number (int): The CC number to use for the connection
            min_range (float): The minimum parameter value used for the connection
            max_range (float): The maximum parameter value used for the connection
            relative_mode (bool): Whether the parameter changes relative to a previous value

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.ConnectCCToParameter(self._sushi_proto.MidiCCConnection(parameter=self._sushi_proto.ParameterIdentifier(processor_id=processor_id, parameter_id=parameter_id),
                                                                               channel=self._sushi_proto.MidiChannel(channel=channel),
                                                                               port=port,
                                                                               cc_number=cc_number,
                                                                               min_range=min_range,
                                                                               max_range=max_range,
                                                                               relative_mode=relative_mode))
            return self._parent.validate_command_response(cr, f"Connect CC{cc_number} to parameter {parameter_id}")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, f"With MidiCCConnection: parameter: {parameter_id}, channel: {channel}, "
                                               f"port: {port}, cc_number: {cc_number}, min_range: {min_range}, "
                                               f"max_range: {max_range}, relative_mode: {relative_mode}.")

    def connect_pc_to_processor(self, processor: int, channel: int, port: int) -> SushiCommandResponse:
        """
        Connects MIDI Program Change messages to a processor

        Parameters:
            processor (int): The id of the processor to connect
            channel (int): The MIDI channel to use for the connection
            port (int): The MIDI port to use for the connection

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.ConnectPCToProcessor(self._sushi_proto.MidiPCConnection(processor=self._sushi_proto.ProcessorIdentifier(id=processor),
                                                                               channel=self._sushi_proto.MidiChannel(channel=channel),
                                                                               port=port))
            return self._parent.validate_command_response(cr, f"Connect PC to processor {processor}")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, f"With MidiPCConnection: processor: {processor}, channel: {channel}, "
                                               f"port: {port}.")

    def disconnect_kbd_input(self, track: int, channel: int, port: int, raw_midi: bool) -> SushiCommandResponse:
        """
        Disconnects a MIDI Keyboard input connection from a track

        Parameters:
            track (int): The id of the track to disconnect
            channel (int): The MIDI channel to disconnect
            port (int): The MIDI port to disconnect
            raw_midi (bool): Disconnect raw MIDI

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.DisconnectKbdInput(self._sushi_proto.MidiKbdConnection(track=self._sushi_proto.TrackIdentifier(id=track),
                                                                              channel=self._sushi_proto.MidiChannel(channel=channel),
                                                                              port=port,
                                                                              raw_midi=raw_midi))
            return self._parent.validate_command_response(cr, f"Disconnect Keyboard input from track {track}")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, f"With MidiKbdConnection: track: {track}, channel: {channel}, "
                                               f"port: {port}, raw_midi: {raw_midi}.")

    def disconnect_kbd_output(self, track: int, channel: int, port: int, raw_midi: bool) -> SushiCommandResponse:
        """
        Disconnects a MIDI Keyboard output connection from a track

        Parameters:
            track (int): The id of the track to disconnect
            channel (int): The MIDI channel to disconnect
            port (int): The MIDI port to disconnect
            raw_midi (bool): Disconnect raw MIDI

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.DisconnectKbdOutput(self._sushi_proto.MidiKbdConnection(track=self._sushi_proto.TrackIdentifier(id=track),
                                                                               channel=self._sushi_proto.MidiChannel(channel=channel),
                                                                               port=port,
                                                                               raw_midi=raw_midi))
            return self._parent.validate_command_response(cr, f"Disconnect MIDI output from track {track}")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, f"With MidiKbdConnection: track: {track}, channel: {channel}, "
                                               f"port: {port}, raw_midi: {raw_midi}.")

    def disconnect_cc(self, processor_id: int, parameter_id: int, channel: int, port: int, cc_number: int,
                                min_range: float, max_range: float, relative_mode: bool) -> SushiCommandResponse:
        """
        Disconnects a MIDI Control Change connection

        Parameters:
            processor_id (int): The id of the processor the parameter belongs to
            parameter (int): The id of the parameter to connect to
            channel (int): The MIDI channel to use for the connection
            port (int): The id of the MIDI port to use for the connection
            cc_number (int): The cc number to use for the connection
            min_range (float): The minimum parameter value used for the connection
            max_range (float): The maximum parameter value used for the connection
            relative_mode (bool): Whether the parameter changes relative to a previous value

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.DisconnectCC(self._sushi_proto.MidiCCConnection(parameter=self._sushi_proto.ParameterIdentifier(processor_id=processor_id, parameter_id=parameter_id),
                                                                       channel=self._sushi_proto.MidiChannel(channel=channel),
                                                                       port=port,
                                                                       cc_number=cc_number,
                                                                       min_range=min_range,
                                                                       max_range=max_range,
                                                                       relative_mode=relative_mode))
            return self._parent.validate_command_response(cr, f"Disconnect CC from parameter {parameter_id}")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, f"With MidiCCConnection: parameter: {parameter_id}, channel: {channel}, "
                                               f"port: {port}, cc_number: {cc_number}, min_range: {min_range}, "
                                               f"max_range: {max_range}, relative_mode: {relative_mode}.")

    def disconnect_pc(self, processor: int, channel: int, port: int) -> SushiCommandResponse:
        """
        Disconnects a MIDI Program Change connection

        Parameters:
            processor (int): The id of the processor to connect
            channel (int): The MIDI channel to use for the connection
            port (int): The MIDI port to use for the connection

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.DisconnectPC(self._sushi_proto.MidiPCConnection(processor=self._sushi_proto.ProcessorIdentifier(id=processor),
                                                                       channel=self._sushi_proto.MidiChannel(channel=channel),
                                                                       port=port))
            return self._parent.validate_command_response(cr, f"Disconnect PC from processor {processor}")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, f"With MidiPCConnection: processor: {processor}, channel: {channel}, "
                                               f"port: {port}.")

    def disconnect_all_cc_from_processor(self, processor_id: int) -> SushiCommandResponse:
        """
        Disconnects all MIDI Control Change connections from a processor

        Parameters:
            processor_id (int): The id of the processor to disconnect

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.DisconnectAllCCFromProcessor(self._sushi_proto.ProcessorIdentifier(id=processor_id))
            return self._parent.validate_command_response(cr, f"Disconnect all CC from processor {processor_id}")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, f"With processor_id: {processor_id}")

    def disconnect_all_pc_from_processor(self, processor_id: int) -> SushiCommandResponse:
        """
        Disconnects all MIDI Program Change connections from a processor

        Parameters:
            processor_id (int): The id of processor to disconnect

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.DisconnectAllPCFromProcessor(self._sushi_proto.ProcessorIdentifier(id=processor_id))
            return self._parent.validate_command_response(cr, f"Disconnect all PC from processor {processor_id}")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, f"With processor_id: {processor_id}")



import asyncio
from .sushierrors import SushiUnkownError


class SushiCommandResponse(asyncio.Event):
    command_id: int
    error: bool = False

    def __init__(self, id: int, result=None) -> None:
        self.id = id
        self.result = result
        super().__init__()

    async def wait(self):
        r = await super().wait()
        if self.error:
            raise SushiUnkownError('Sushi returned an error')
        return r



__author__ = "Ruben Svensson"
__copyright__ = """

    Copyright 2017-2019 Modern Ancient Instruments Networked AB, dba Elk

    elkpy is free software: you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation, either version 3
    of the License, or (at your option) any later version.

    elkpy is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with elkpy.  If
    not, see <http://www.gnu.org/licenses/>.
"""
__license__ = "GPL-3.0"

from pathlib import Path
from typing import Any
import re
import sys

from .events import SushiCommandResponse

from . import audiographcontroller
from . import keyboardcontroller
from . import parametercontroller
from . import programcontroller
from . import timingcontroller
from . import transportcontroller
from . import audioroutingcontroller
from . import midicontroller
from . import cvgatecontroller
from . import osccontroller
from . import systemcontroller
from . import sessioncontroller
from . import notificationcontroller
from . import grpc_gen

from .sushi_info_types import CommandResponseStatus
from . import sushierrors


sushi_proto = Path(__file__).parent / "rpc" / "sushi_rpc.proto"


###############################
# Main sushi controller class #
###############################


class SushiController:
    """
    A class to control sushi via gRPC.
    This class creates one instance of each different controller type and makes
    these sub-controllers available as member variables. See the documentation
    of the separate sub-controllers for their usage.

    Attributes:
        _stub (SushiControllerStub): Connection stubs to the gRPC interface implemented in sushi.

    Notes:
        close() should ALWAYS be called as part of an application housekeeping/cleanup-before-shutdown routine as it
        ensure proper releasing of resources and clean joining of concurrent threads.
    """

    def __init__(
        self,
        address="localhost:51051",
        sushi_proto_def=sushi_proto,
        auto_update_async_command_responses: bool = True,
    ):
        """
        The constructor for the SushiController class setting up the gRPC connection with sushi.

        Parameters:
            - address (str): 'ip-addres:port' The ip-addres and port at which to connect to sushi.
            - sushi_proto_def (str): path to .proto file with SUSHI's gRPC services definition
            - auto_update_async_command_responses (bool): controls whether the notificaton controller
                    listens to async command response notifications and manages pending command
                    responses.
        """
        self._sushi_proto, self._sushi_grpc = grpc_gen.modules_from_proto(
            str(sushi_proto_def)
        )

        self.audio_graph = audiographcontroller.AudioGraphController(
            self, address, str(sushi_proto_def), self._sushi_proto, self._sushi_grpc
        )
        self.keyboard = keyboardcontroller.KeyboardController(
            self, address, str(sushi_proto_def), self._sushi_proto, self._sushi_grpc
        )
        self.parameters = parametercontroller.ParameterController(
            self, address, str(sushi_proto_def), self._sushi_proto, self._sushi_grpc
        )
        self.programs = programcontroller.ProgramController(
            self, address, str(sushi_proto_def)
        )
        self.timings = timingcontroller.TimingController(
            self, address, str(sushi_proto_def), self._sushi_proto, self._sushi_grpc
        )
        self.transport = transportcontroller.TransportController(
            self, address, str(sushi_proto_def), self._sushi_proto, self._sushi_grpc
        )
        self.audio_routing = audioroutingcontroller.AudioRoutingController(
            self, address, str(sushi_proto_def), self._sushi_proto, self._sushi_grpc
        )
        self.midi_controller = midicontroller.MidiController(
            self, address, str(sushi_proto_def), self._sushi_proto, self._sushi_grpc
        )
        self.cv_gate_controller = cvgatecontroller.CvGateController(
            self, address, str(sushi_proto_def), self._sushi_proto, self._sushi_grpc
        )
        self.osc_controller = osccontroller.OscController(
            self, address, str(sushi_proto_def), self._sushi_proto, self._sushi_grpc
        )
        self.system = systemcontroller.SystemController(address, str(sushi_proto_def))
        self.session = sessioncontroller.SessionController(
            self, address, str(sushi_proto_def), self._sushi_proto, self._sushi_grpc
        )

        if not self.check_sushi_api_compatibility():
            sys.exit()

        self.notifications = notificationcontroller.NotificationController(
            self,
            address,
            str(sushi_proto_def),
            self._sushi_proto,
            self._sushi_grpc,
            auto_update_async_command_responses,
        )

        self.command_responses: dict[int, SushiCommandResponse] = {}

    def check_sushi_api_compatibility(self) -> bool:
        """Verifies that the version of the api used by elkpy matches the one build in Sushi"""

        api_proto_file = Path(__file__).parent / "rpc" / "api_version.proto"

        with open(api_proto_file, 'r') as f:
            content = f.read()
        
        # Find: [default = "something"]
        match = re.search(r'\[default\s*=\s*"([^"]+)"\]', content)
        
        if match:
            api_version = match.group(1)
            sushi_version = self.system.get_sushi_api_version() 
            if api_version == sushi_version:
                print("API match!")
                return True
            else:
                print(f"API version mismatch! Sushi = {sushi_version} <-> Elkpy = {api_version}")
                sys.exit()
        
        raise ValueError("No default api version found")


    def validate_command_response(self, cr, context_info: str = "") -> Any:
        match CommandResponseStatus(cr.status.status):
            case CommandResponseStatus.SUCCESS:
                return cr
            case CommandResponseStatus.ERROR:
                raise sushierrors.SushiUnkownError(context_info)
            case CommandResponseStatus.UNSUPPORTED_OPERATION:
                raise sushierrors.SushiUnsupportedOperationError(context_info)
            case CommandResponseStatus.NOT_FOUND:
                raise sushierrors.SushiNotFoundError(context_info)
            case CommandResponseStatus.OUT_OF_RANGE:
                raise sushierrors.SushiOutOfRangeError(context_info)
            case CommandResponseStatus.INVALID_ARGUMENTS:
                raise sushierrors.SushiInvalidArgumentError(context_info)
            case CommandResponseStatus.ASYNC_RESPONSE:
                event = SushiCommandResponse(id=cr.id)
                self.command_responses[cr.id] = event
                return event

    def close(self):
        """
        This method should be called at app close.
        It should call any sub-controller close routines whenever they exist.
        i.e.: NotificationController has an infinite event loop running in its own thread, which has to be stopped and joined
        to ensure clean closing and proper releasing of any resources.
        """
        self.notifications.close()

    def __del__(self):
        self.notifications.close()



__author__ = "Ruben Svensson"
__copyright__ = """

    Copyright 2017-2019 Modern Ancient Instruments Networked AB, dba Elk

    elkpy is free software: you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation, either version 3
    of the License, or (at your option) any later version.

    elkpy is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with elkpy.  If
    not, see <http://www.gnu.org/licenses/>.
"""
__license__ = "GPL-3.0"

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from elkpy.sushicontroller import SushiController

from .events import SushiCommandResponse
import grpc

from . import sushierrors
from . import grpc_gen
from . import sushi_info_types as info_types
from typing import List

####################################
# Sushi parameter controller class #
####################################

class ParameterController:
    """
    A class to control the parameter in sushi via gRPC. It manages the parameters of sushi. Enabling
    getting and setting of values as well as getting info about what parameters are availble and their
    ranges, types, etc.

    Attributes:
        _stub (ParameterControllerStub): Connection stubs to the gRPC parameter interface implemented in sushi.
    """
    def __init__(self,
                 parent: 'SushiController',
                 address = 'localhost:51051',
                 sushi_proto_def = '/usr/share/sushi/sushi_rpc.proto',
                 sushi_proto = None, sushi_grpc = None):
        """
        The constructor for the ParameterController class setting up the gRPC connection with sushi.

        Parameters:
            address (str): 'ip-addres:port' The ip-addres and port at which to connect to sushi.
            sushi_proto_def (str): path to .proto file with SUSHI's gRPC services definition
        """
        try:
            channel = grpc.insecure_channel(address)
        except AttributeError as e:
            raise TypeError("Parameter address = {}. Should be a string containing the ip-address and port of sushi ('ip-address:port')".format(address)) from e

        if sushi_proto and sushi_grpc:
            self._sushi_proto = sushi_proto
            self._sushi_grpc = sushi_grpc
        else:
            self._sushi_proto, self._sushi_grpc = grpc_gen.modules_from_proto(sushi_proto_def)

        self._stub = self._sushi_grpc.ParameterControllerStub(channel)
        self._parent = parent

    def get_track_parameters(self, track_identifier: int) -> List[info_types.ParameterInfo]:
        """
        Get a list of parameters available on the specified track.

        Parameters:
            track_identifier (int): The id of the track to get the parameter list from.

        Returns:
            List[info_types.ParameterInfo]: A list of the info of the parameters assigned to the track matching the id.
        """
        try:
            cr = self._stub.GetTrackParameters(self._sushi_proto.TrackIdentifier(
                id = track_identifier
            ))

            response = self._parent.validate_command_response(cr, f"Get parameters for track {track_identifier}")
            parameter_info_list = []
            for parameter_info in response.parameters:
                parameter_info_list.append(info_types.ParameterInfo(parameter_info))

            return parameter_info_list

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, "With track id: {}".format(track_identifier))

    def get_processor_parameters(self, processor_identifier: int) -> List[info_types.ParameterInfo]:
        """
        Get a list of the parameters available to the specified processor.

        Parameters:
            processor_identifier (int): The id of the processor to get the parameters from.

        Returns:
            List[info_types.ParameterInfo]: A list of the parameters available to the processor matching the id.
        """
        try:
            cr = self._stub.GetProcessorParameters(self._sushi_proto.ProcessorIdentifier(
                id = processor_identifier
            ))
            response = self._parent.validate_command_response(cr, f"Get parameters for processor {processor_identifier}")

            parameter_info_list = []
            for parameter_info in response.parameters:
                parameter_info_list.append(info_types.ParameterInfo(parameter_info))

            return parameter_info_list

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, "With processor id: {}".format(processor_identifier))

    def get_parameter_id(self, processor_identifier: int, parameter_name: str) -> int:
        """
        Get the id of the parameter of the specified processor corresponding to the specified parameter name.

        Parameters:
            processor_identifier (int): The id of the processor to get the parameter id from.
            parameter_name (str): The name of the parameter to get the id from.

        Returns:
            int: The id of the parameter matching the parameter name.
        """
        try:
            cr = self._stub.GetParameterId(self._sushi_proto.ParameterIdRequest(
                processor = self._sushi_proto.ProcessorIdentifier(id = processor_identifier),
                ParameterName = parameter_name
            ))
            response = self._parent.validate_command_response(cr, f"Get parameters for processor {processor_identifier}")
            return response.id.parameter_id

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, "With processor id: {}, parameter name: {}".format(processor_identifier, parameter_name))

    def get_parameter_info(self, processor_identifier: int, parameter_identifier: int) -> info_types.ParameterInfo:
        """
        Get info about the specified parameter on the specified processor.

        Parameters:
            processor_identifier (int): The id of the processor to get the parameter info from.
            parameter_identifier (int): The id of the parameter to get the info from.

        Returns:
            info_types.ParameterInfo: Info of the parameter matching the id.
        """
        try:
            cr = self._stub.GetParameterInfo(self._sushi_proto.ParameterIdentifier(
                processor_id = processor_identifier,
                parameter_id = parameter_identifier
            ))
            response = self._parent.validate_command_response(cr, f"Get info for parameter {parameter_identifier}")
            return info_types.ParameterInfo(response.info)

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, "With processor id: {}, parameter id: {}".format(processor_identifier, parameter_identifier))

    def get_parameter_value(self, processor_identifier: int, parameter_identifier: int) -> float:
        """
        Get the value of the parameter matching the specified parameter on the specified processor.

        Parameters:
            processor_identifier (int): The id of the processor to get the parameter value from.
            parameter_identifier (int): The id of the parameter to get the value from.

        Returns:
            float: The value of the parameter matching the id.
        """
        try:
            cr = self._stub.GetParameterValue(self._sushi_proto.ParameterIdentifier(
                processor_id = processor_identifier,
                parameter_id = parameter_identifier
            ))
            response = self._parent.validate_command_response(cr, f"Get value of parameter {parameter_identifier}")
            return response.value

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, "With processor id: {}, parameter id: {}".format(processor_identifier, parameter_identifier))

    def get_parameter_value_in_domain(self, processor_identifier: int, parameter_identifier: int) -> float:
        """
        Get the normalised value of the parameter matching the specified parameter on the specified processor.

        Parameters:
            processor_identifier (int): The id of the processor to get the normalised parameter value from.
            parameter_identifier (int): The id of the parameter to get the normalised value from.

        Returns:
            float: The normalised value of the parameter matching the id.
        """
        try:
            cr = self._stub.GetParameterValueInDomain(self._sushi_proto.ParameterIdentifier(
                processor_id = processor_identifier,
                parameter_id = parameter_identifier
            ))
            response = self._parent.validate_command_response(cr, f"Get value of parameter {parameter_identifier} in domain")
            return response.value

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, "With processor id: {}, parameter id: {}".format(processor_identifier,parameter_identifier))

    def get_parameter_value_as_string(self, processor_identifier: int, parameter_identifier: int) -> str:
        """
        Get the value of the parameter matching the specified parameter on the specified processor as a string.

        Parameters:
            processor_identifier (int): The id of the processor to get the parameter value string from.
            parameter_identifier (int): The id of the parameter to get value string from.

        Returns:
            str: The value as a string of the parameter matching the id.
        """
        try:
            cr = self._stub.GetParameterValueAsString(self._sushi_proto.ParameterIdentifier(
                processor_id = processor_identifier,
                parameter_id = parameter_identifier
            ))
            response = self._parent.validate_command_response(cr, f"Get value of parameter {parameter_identifier} as string")
            return response.value

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, "With processor id: {}, parameter id: {}".format(processor_identifier, parameter_identifier))

    def set_parameter_value(self, processor_identifier: int, parameter_identifier: int, value: float) -> SushiCommandResponse:
        """
        Set the value of the specified parameter on the specified processor.

        Parameters:
            processor_identifier (int): The id of the processor that has the parameter to be changed.
            parameter_identifier (int): The id of the property to set the value of.

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.SetParameterValue(self._sushi_proto.ParameterValue(
                parameter = self._sushi_proto.ParameterIdentifier(
                    processor_id = processor_identifier,
                    parameter_id = parameter_identifier
                    ),
                value = value
            ))
            return self._parent.validate_command_response(cr, f"Get value of parameter {parameter_identifier} as string")

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, "With processor id: {}, property id: {}, value: {}".format(processor_identifier, parameter_identifier, value))

    def get_track_properties(self, track_identifier: int) -> List[info_types.PropertyInfo]:
        """
        Get a list of string properties available on the specified track.

        Parameters:
            track_identifier (int): The id of the track to get the property list from.

        Returns:
            List[info_types.PropertyInfo]: A list of the info of the properties assigned to the track matching the id.
        """
        try:
            cr = self._stub.GetTrackProperties(self._sushi_proto.TrackIdentifier(
                id = track_identifier
            ))
            response = self._parent.validate_command_response(cr, f"Get properties for track {track_identifier}")

            property_info_list = []
            for property_info in response.properties:
                property_info_list.append(info_types.PropertyInfo(property_info))

            return property_info_list

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, "With track id: {}".format(track_identifier))

    def get_processor_properties(self, processor_identifier: int) -> List[info_types.PropertyInfo]:
        """
        Get a list of the string properties available to the specified processor.

        Parameters:
            processor_identifier (int): The id of the processor to get the properties from.

        Returns:
            List[info_types.PropertyInfo]: A list of the properties available to the processor matching the id.
        """
        try:
            cr = self._stub.GetProcessorProperties(self._sushi_proto.ProcessorIdentifier(
                id = processor_identifier
            ))
            response = self._parent.validate_command_response(cr, f"Get properties for processor {processor_identifier}")

            property_info_list = []
            for property_info in response.properties:
                property_info_list.append(info_types.PropertyInfo(property_info))

            return property_info_list

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, "With processor id: {}".format(processor_identifier))

    def get_property_id(self, processor_identifier: int, property_name: str) -> int:
        """
        Get the id of the property of the specified processor corresponding to the specified property name.

        Parameters:
            processor_identifier (int): The id of the processor to get the parameter id from.
            property_name (str): The name of the property to get the id from.

        Returns:
            int: The id of the property matching the property name.
        """
        try:
            cr = self._stub.GetPropertyId(self._sushi_proto.PropertyIdRequest(
                processor = self._sushi_proto.ProcessorIdentifier(id = processor_identifier),
                property_name = property_name
            ))
            response = self._parent.validate_command_response(cr, f"Get id for property {property_name}")
            return response.id.property_id

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, "With processor id: {}, string property name: {}".format(processor_identifier, property_name))

    def get_property_info(self, processor_identifier: int, property_identifier: int) -> info_types.PropertyInfo:
        """
        Get info about the specified property on the specified processor.

        Parameters:
            processor_identifier (int): The id of the processor to get the parameter info from.
            property_identifier (int): The id of the string property to get the info from.

        Returns:
            info_types.PropertyInfo: Info of the property matching the id.
        """
        try:
            cr = self._stub.GetPropertyInfo(self._sushi_proto.PropertyIdentifier(
                processor_id = processor_identifier,
                property_id = property_identifier
            ))
            response = self._parent.validate_command_response(cr, f"Get info for property {processor_identifier}")
            return info_types.PropertyInfo(response.info)

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, "With processor id: {}, property id: {}".format(processor_identifier, property_identifier))

    def get_property_value(self, processor_identifier: int, property_identifier: int) -> str:
        """
        Get the value of the property matching the specified property on the specified processor.

        Parameters:
            processor_identifier (int): The id of the processor to get the parameter value string from.
            property_identifier (int): The id of the string property to get value string from.

        Returns:
            str: The value of the property matching the id.
        """
        try:
            cr = self._stub.GetPropertyValue(self._sushi_proto.PropertyIdentifier(
                processor_id = processor_identifier,
                property_id = property_identifier
            ))
            response = self._parent.validate_command_response(cr, f"Get value of property {property_identifier}")
            return response.value

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, "With processor id: {}, property id: {}".format(processor_identifier, property_identifier))

    def set_property_value(self, processor_identifier: int, property_identifier: int, value: float) -> SushiCommandResponse:
        """
        Set the value of the specified property on the specified processor.

        Parameters:
            processor_identifier (int): The id of the processor that has the property to be changed.
            property_identifier (int): The id of the property to set the value of.
            value (string) : The new value to assign to the property

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.SetPropertyValue(self._sushi_proto.PropertyValue(
                property = self._sushi_proto.PropertyIdentifier(
                    processor_id = processor_identifier,
                    property_id = property_identifier
                    ),
                value = value
            ))
            return self._parent.validate_command_response(cr, f"Set value of property {property_identifier}")

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, "With processor id: {}, property id: {}, value: {}".format(processor_identifier, property_identifier, value))



__author__ = "Maxime Gendebien"
__copyright__ = """

    Copyright 2017-2019 Modern Ancient Instruments Networked AB, dba Elk

    elkpy is free software: you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation, either version 3
    of the License, or (at your option) any later version.

    elkpy is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with elkpy.  If
    not, see <http://www.gnu.org/licenses/>.
"""
__license__ = "GPL-3.0"

from typing import TYPE_CHECKING, List
if TYPE_CHECKING:
    from elkpy.sushicontroller import SushiController

from .events import SushiCommandResponse
import grpc
from . import sushierrors
from . import sushi_info_types as info_types
from . import grpc_gen

######################################
#   Sushi CV-Gate Controller class   #
######################################


class CvGateController:
    """
    Class to manage CV/Gate connections in Sushi via gRPC. It creates, deletes and gets info about
    Cv/Gate connections on sushi tracks.

    Attributes:
        _stub (CvGateControllerStub): Connection stub to the gRPC CvGate interface.
    """
    def __init__(self,
                 parent: 'SushiController',
                 address='localhost:51051',
                 sushi_proto_def='/usr/share/sushi/sushi_rpc.proto',
                 sushi_proto = None, sushi_grpc = None) -> None:
        """
        The constructor for the CvGateController class setting up the gRPC connection with sushi.

        Args:
            address: IP address to Sushi in the uri form : 'ip-addr:port'
            sushi_proto_def: path to the .proto file with SUSHI gRPC services definitions
        """
        try:
            channel = grpc.insecure_channel(address)
        except AttributeError as e:
            raise TypeError(f"Parameter address = {address}. Should be a string with the IP address and port of Sushi") from e

        if sushi_proto and sushi_grpc:
            self._sushi_proto = sushi_proto
            self._sushi_grpc = sushi_grpc
        else:
            self._sushi_proto, self._sushi_grpc = grpc_gen.modules_from_proto(sushi_proto_def)

        self._stub = self._sushi_grpc.CvGateControllerStub(channel)
        self._parent = parent

    def get_cv_input_channel_count(self) -> int:
        """
        Gets a count of all CV input channels

        Returns:
            int: count
        """
        try:
            response = self._stub.GetCvInputChannelCount(self._sushi_proto.GenericVoidValue())
            return response.value
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def get_cv_output_channel_count(self) -> int:
        """
        Gets a count of all CV output channels

        Returns:
            int: count
        """
        try:
            response = self._stub.GetCvOutputChannelCount(self._sushi_proto.GenericVoidValue())
            return response.value
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def get_all_cv_input_connections(self) -> List[info_types.CvConnection]:
        """
        Gets all CV input connections

        Returns:
            List[info_types.CvConnection]: List of CvConnection objects
        """
        try:
            response = self._stub.GetAllCvInputConnections(self._sushi_proto.GenericVoidValue())
            return [info_types.CvConnection(c) for c in response.connections]
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def get_all_cv_output_connections(self) -> List[info_types.CvConnection]:
        """
        Gets all CV output connections

        Returns:
            List[info_types.CvConnection]: List of CvConnection objects
        """
        try:
            response = self._stub.GetAllCvOutputConnections(self._sushi_proto.GenericVoidValue())
            return [info_types.CvConnection(c) for c in response.connections]
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def get_all_gate_input_connections(self) -> List[info_types.GateConnection]:
        """
        Gets all Gate input connections

        Returns:
            List[info_types.GateConnection]: List of GateConnection objects
        """
        try:
            response = self._stub.GetAllGateInputConnections(self._sushi_proto.GenericVoidValue())
            return [info_types.GateConnection(c) for c in response.connections]
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def get_all_gate_output_connections(self) -> List[info_types.GateConnection]:
        """
        Gets all Gate output connections

        Returns:
            List[info_types.GateConnection]: List of GateConnection objects
        """
        try:
            response = self._stub.GetAllGateOutputConnections(self._sushi_proto.GenericVoidValue())
            return [info_types.GateConnection(c) for c in response.connections]
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def get_cv_input_connections_for_processor(self, processor_id: int) -> List[info_types.CvConnection]:
        """
        Gets a list of all CV input connections for specified processor.

        Returns:
            List[info_types.CvConnection]: List of CvConnection objects
        """
        try:
            cr = self._stub.GetCvInputConnectionsForProcessor(self._sushi_proto.ProcessorIdentifier(id=processor_id))
            response = self._parent.validate_command_response(cr, f"Get CV input connections for processor {processor_id}")
            return [info_types.CvConnection(c) for c in response.connections]
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def get_cv_output_connections_for_processor(self, processor_id: int) -> List[info_types.CvConnection]:
        """
        Gets a list of all CV output connections for specified processor.

        Returns:
            List[info_types.CvConnection]: List of CvConnection objects
        """
        try:
            cr = self._stub.GetCvOutputConnectionsForProcessor(self._sushi_proto.ProcessorIdentifier(id=processor_id))
            response = self._parent.validate_command_response(cr, f"Get CV output connections for processor {processor_id}")
            return [info_types.CvConnection(c) for c in response.connections]
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def get_gate_input_connections_for_processor(self, processor_id: int) -> List[info_types.GateConnection]:
        """
        Gets a list of all Gate input connections for specified processor.

        Returns:
            List[info_types.GateConnection]: List of GateConnection objects
        """
        try:
            cr = self._stub.GetGateInputConnectionsForProcessor(self._sushi_proto.ProcessorIdentifier(id=processor_id))
            response = self._parent.validate_command_response(cr, f"Get gate input connections for processor {processor_id}")
            return [info_types.GateConnection(c) for c in response.connections]
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def get_gate_output_connections_for_processor(self, processor_id: int) -> List[info_types.GateConnection]:
        """
        Gets a list of all Gate output connections for specified processor.

        Returns:
            List[info_types.GateConnection]: List of GateConnection objects
        """
        try:
            cr = self._stub.GetGateOutputConnectionsForProcessor(self._sushi_proto.ProcessorIdentifier(id=processor_id))
            response = self._parent.validate_command_response(cr, f"Get gate output connections for processor {processor_id}")
            return [info_types.GateConnection(c) for c in response.connections]
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def connect_cv_input_to_parameter(self, processor_id: int, parameter_id: int, cv_port_id: int) -> SushiCommandResponse:
        """
        Connects a CV input to a parameter

        Parameters:
            processor_id (int): The id of the processor to parameter belongs to
            parameter_id (int): The id of parameter to connect to
            cv_port_id (int): The id of the CV port to connect to

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.ConnectCvInputToParameter(self._sushi_proto.CvConnection(parameter=self._sushi_proto.ParameterIdentifier(processor_id=processor_id, parameter_id=parameter_id),
                                                                                cv_port_id=cv_port_id))
            return self._parent.validate_command_response(cr, f"Connect CV input to parameter {parameter_id}")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def connect_cv_output_from_parameter(self, processor_id: int, parameter_id: int, cv_port_id: int) -> SushiCommandResponse:
        """
        Connects a CV output to a parameter

        Parameters:
            processor_id (int): The id of the processor to parameter belongs to
            parameter_id (int): The id of parameter to connect to
            cv_port_id (int): The id of the CV port to connect to

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.ConnectCvOutputFromParameter(self._sushi_proto.CvConnection(parameter=self._sushi_proto.ParameterIdentifier(processor_id=processor_id, parameter_id=parameter_id),
                                                                                   cv_port_id=cv_port_id))
            return self._parent.validate_command_response(cr, f"Connect CV output to parameter {parameter_id}")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def connect_gate_input_to_processor(self, processor: int, gate_port_id: int, channel: int, note_no: int) -> SushiCommandResponse:
        """
        Connects a Gate input to a processor

        Parameters:
            processor (int): The id of the processor to connect to
            gate_port_id (int): The id of the gate port being connected
            channel (int): The midi channel number
            note_no (int): The note number to trigger

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.ConnectGateInputToProcessor(self._sushi_proto.GateConnection(processor=self._sushi_proto.ProcessorIdentifier(id=processor),
                                                                                    gate_port_id=gate_port_id,
                                                                                    channel=channel,
                                                                                    note_no=note_no))
            return self._parent.validate_command_response(cr, f"Connect gate input to processor {processor}")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def connect_gate_output_from_processor(self, processor: int, gate_port_id: int, channel: int, note_no: int) -> SushiCommandResponse:
        """
        Connects a Gate output from a processor

        Parameters:
            processor (int): The id of the processor to connect to
            gate_port_id (int): The id of the gate port being connected
            channel (int): The midi channel number
            note_no (int): The note number to trigger

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.ConnectGateOutputFromProcessor(self._sushi_proto.GateConnection(processor=self._sushi_proto.ProcessorIdentifier(id=processor),
                                                                                       gate_port_id=gate_port_id,
                                                                                       channel=channel,
                                                                                       note_no=note_no))
            return self._parent.validate_command_response(cr, f"Connect gate output from processor {processor}")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def disconnect_cv_input(self, processor_id: int, parameter_id: int, cv_port_id: int) -> SushiCommandResponse:
        """
        Disconnects a CV input from a parameter

        Parameters:
            procsessor_id (int): The id of the processor the parameter belongs to
            parameter_id (int): The id of the parameter to disconnect
            cv_port_id (int): The id of the CV port to disconnect

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.DisconnectCvInput(self._sushi_proto.CvConnection(parameter=self._sushi_proto.ParameterIdentifier(processor_id=processor_id, parameter_id=parameter_id),
                                                                        cv_port_id=cv_port_id))
            return self._parent.validate_command_response(cr, f"Disconnect CV input from processor {processor_id}")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def disconnect_cv_output(self, processor_id: int, parameter_id: int, cv_port_id: int) -> SushiCommandResponse:
        """
        Disconnects a CV output from a parameter

        Parameters:
            procsessor_id (int): The id of the processor the parameter belongs to
            parameter_id (int): The id of the parameter to disconnect
            cv_port_id (int): The id of the CV port to disconnect

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.DisconnectCvOutput(self._sushi_proto.CvConnection(parameter=self._sushi_proto.ParameterIdentifier(processor_id=processor_id, parameter_id=parameter_id),
                                                                         cv_port_id=cv_port_id))
            return self._parent.validate_command_response(cr, f"Disconnect CV output from processor {processor_id}")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def disconnect_gate_input(self, processor: int, gate_port_id: int, channel: int, note_no: int) -> SushiCommandResponse:
        """
        Disconnects a Gate input from a processor

        Parameters:
            processor (int): The id of the processor to connect to
            gate_port_id (int): The id of the gate port being connected
            channel (int): The midi channel number
            note_no (int): The note number to trigger

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.DisconnectGateInput(self._sushi_proto.GateConnection(processor=self._sushi_proto.ProcessorIdentifier(id=processor),
                                                                            gate_port_id=gate_port_id,
                                                                            channel=channel,
                                                                            note_no=note_no))
            return self._parent.validate_command_response(cr, f"Disconnect gate input from processor {processor}")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def disconnect_gate_output(self, processor: int, gate_port_id: int, channel: int, note_no: int) -> SushiCommandResponse:
        """
        Disconnects a Gate output from a processor

        Parameters:
            processor (int): The id of the processor to connect to
            gate_port_id (int): The id of the gate port being connected
            channel (int): The midi channel number
            note_no (int): The note number to trigger

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.DisconnectGateOutput(self._sushi_proto.GateConnection(processor=self._sushi_proto.ProcessorIdentifier(id=processor),
                                                                             gate_port_id=gate_port_id,
                                                                             channel=channel,
                                                                             note_no=note_no))
            return self._parent.validate_command_response(cr, f"Disconnect gate output from processor {processor}")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def disconnect_all_cv_inputs_from_processor(self, processor: int) -> SushiCommandResponse:
        """
        Disconnects all CV inputs from a processor.

        Parameters:
            processor (int): the processor ID from which to disconnect CV inputs

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.DisconnectAllCvInputsFromProcessor(self._sushi_proto.ProcessorIdentifier(id=processor))
            return self._parent.validate_command_response(cr, f"Disconnect all CV inputs from processor {processor}")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def disconnect_all_cv_outputs_from_processor(self, processor: int) -> SushiCommandResponse:
        """
        Disconnects all CV outputs from a processor.

        Parameters:
            processor (int): the processor ID from which to disconnect CV outputs

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.DisconnectAllCvOutputsFromProcessor(self._sushi_proto.ProcessorIdentifier(id=processor))
            return self._parent.validate_command_response(cr, f"Disconnect all CV outputs from processor {processor}")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def disconnect_all_gate_inputs_from_processor(self, processor: int) -> SushiCommandResponse:
        """
        Disconnects all Gate inputs from a processor.

        Parameters:
            processor (int): the processor ID from which to disconnect Gate inputs

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.DisconnectAllGateInputsFromProcessor(self._sushi_proto.ProcessorIdentifier(id=processor))
            return self._parent.validate_command_response(cr, f"Disconnect all gate inputs from processor {processor}")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def disconnect_all_gate_outputs_from_processor(self, processor: int) -> SushiCommandResponse:
        """
        Disconnects all Gate outputs from a processor.

        Parameters:
            processor (int): the processor ID from which to disconnect Gate outputs

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.DisconnectAllGateOutputsFromProcessor(self._sushi_proto.ProcessorIdentifier(id=processor))
            return self._parent.validate_command_response(cr, f"Disconnect all gate outputs from processor {processor}")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)



__author__ = "Ruben Svensson"
__copyright__ = """

    Copyright 2017-2019 Modern Ancient Instruments Networked AB, dba Elk

    elkpy is free software: you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation, either version 3
    of the License, or (at your option) any later version.

    elkpy is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with elkpy.  If
    not, see <http://www.gnu.org/licenses/>.
"""
__license__ = "GPL-3.0"

import grpc

from . import sushierrors
from . import grpc_gen

####################################
# Sushi keyboard controller class #
####################################

class KeyboardController(object):
    """
    A class to control the keyboard in sushi via gRPC. It can send typical keyboard events like
    note on, note off, pitch bend, modulation and after touch.

    Attributes:
        _stub (KeyboardControllerStub): Connection stubs to the gRPC keyboard interface implemented in sushi.
    """
    def __init__(self,
                 parent,
                 address = 'localhost:51051',
                 sushi_proto_def = '/usr/share/sushi/sushi_rpc.proto',
                 sushi_proto = None, sushi_grpc = None):
        """
        The constructor for the KeyboardController class setting up the gRPC connection with sushi.

        Parameters:
            address (str): 'ip-addres:port' The ip-addres and port at which to connect to sushi.
            sushi_proto_def (str): path to .proto file with SUSHI's gRPC services definition
        """
        try:
            channel = grpc.insecure_channel(address)
        except AttributeError as e:
            raise TypeError("Parameter address = {}. Should be a string containing the ip-address and port of sushi ('ip-address:port')".format(address)) from e

        if sushi_proto and sushi_grpc:
            self._sushi_proto = sushi_proto
            self._sushi_grpc = sushi_grpc
        else:
            self._sushi_proto, self._sushi_grpc = grpc_gen.modules_from_proto(sushi_proto_def)
        self._stub = self._sushi_grpc.KeyboardControllerStub(channel)
        self._parent = parent

    #######################
    # // Keyboard control #
    #######################

    def send_note_on(self, track_identifier: int, channel: int, note: int, velocity: float) -> None:
        """
        Sends a note on message to the specified track.

        Parameters:
            track_identifier (int): The id of the track that should receive the message.
            channel (int): The channel on which the message should be sent.
            note (int): The note to send. Follows the MIDI standard where middle c = 60.
            velocity (float): The velocity of the note. Should be in range (0.0-1.0).
        """
        try:
            cr = self._stub.SendNoteOn(self._sushi_proto.NoteOnRequest(
                track = self._sushi_proto.TrackIdentifier(id = track_identifier),
                channel = channel,
                note = note,
                velocity = velocity
            ))
            return self._parent.validate_command_response(cr, "Sending NoteOn")

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, " With track id: {}, channel: {}, note: {}, velocity: {}".format(track_identifier, channel, note, velocity))

    def send_note_off(self, track_identifier: int, channel: int, note: int, velocity: float) -> None:
        """
        Sends a note off message to the specified track.

        Parameters:
            track_identifier (int): The id of the track that should receive the message.
            channel (int): The channel on which the message should be sent.
            note (int): The note to send. Follows the MIDI standard where middle c = 60.
            velocity (float): The velocity of the note. Should be in range (0.0-1.0).
        """
        try:
            cr = self._stub.SendNoteOff(self._sushi_proto.NoteOffRequest(
                track = self._sushi_proto.TrackIdentifier(id = track_identifier),
                channel = channel,
                note = note,
                velocity = velocity
            ))
            return self._parent.validate_command_response(cr, "Sending NoteOff")

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, " With track id: {}, channel: {}, note: {}, velocity: {}".format(track_identifier, channel, note, velocity))

    def send_note_aftertouch(self, track_identifier: int, channel: int, note: int, value: float) -> None:
        """
        Sends a aftertouch message to the specified track and note.

        Parameters:
            track_identifier (int): The id of the track that should receive the message.
            channel (int): The channel on which the message should be sent.
            note (int): The note which should receive the message. Follows the MIDI standard where middle c = 60.
            value (float): The aftertouch value of the note. Should be in range (0.0-1.0).
        """
        try:
            cr = self._stub.SendNoteAftertouch(self._sushi_proto.NoteAftertouchRequest(
                track = self._sushi_proto.TrackIdentifier(id = track_identifier),
                channel = channel,
                note = note,
                value = value
            ))
            return self._parent.validate_command_response(cr, "Sending NoteAfterTouch")

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, " With track id: {}, channel: {}, note: {}, value: {}".format(track_identifier, channel, note, value))

    def send_aftertouch(self, track_identifier: int, channel: int, value: float) -> None:
        """
        Sends a aftertouch message to the specified track.

        Parameters:
            track_identifier (int): The id of the track that should receive the message.
            channel (int): The channel on which the message should be sent.
            value (float): The aftertouch value. Should be in range (0.0-1.0).
        """
        try:
            cr = self._stub.SendAftertouch(self._sushi_proto.NoteModulationRequest(
                track = self._sushi_proto.TrackIdentifier(id = track_identifier),
                channel = channel,
                value = value
            ))
            return self._parent.validate_command_response(cr, "Sending AfterTouch")

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, " With track id: {}, channel: {}, value: {}".format(track_identifier, channel, value))

    def send_pitch_bend(self, track_identifier: int, channel: int, value: float) -> None:
        """
        Sends a pitch bend message to the specified track.

        Parameters:
            track_identifier (int): The id of the track that should receive the message.
            channel (int): The channel on which the message should be sent.
            value (float): The pitch bend value. Should be in range (0.0-1.0).
        """
        try:
            cr = self._stub.SendPitchBend(self._sushi_proto.NoteModulationRequest(
                track = self._sushi_proto.TrackIdentifier(id = track_identifier),
                channel = channel,
                value = value
            ))
            return self._parent.validate_command_response(cr, "Sending Pitchbend")

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, " With track id: {}, channel: {}, value: {}".format(track_identifier, channel, value))

    def send_modulation(self, track_identifier: int, channel: int, value: float) -> None:
        """
        Sends a modulation message to the specified track.

        Parameters:
            track_identifier (int): The id of the track that should receive the message.
            channel (int): The channel on which the message should be sent.
            value (float): The modulation value. Should be in range (0.0-1.0).
        """
        try:
            cr = self._stub.SendModulation(self._sushi_proto.NoteModulationRequest(
                track = self._sushi_proto.TrackIdentifier(id = track_identifier),
                channel = channel,
                value = value
            ))
            return self._parent.validate_command_response(cr, "Sending Modulation")

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, " With track id: {}, channel: {}, value: {}".format(track_identifier, channel, value))



__author__ = "Maxime Gendebien"
__copyright__ = """

    Copyright 2017-2019 Modern Ancient Instruments Networked AB, dba Elk

    elkpy is free software: you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation, either version 3
    of the License, or (at your option) any later version.

    elkpy is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with elkpy.  If
    not, see <http://www.gnu.org/licenses/>.
"""
__license__ = "GPL-3.0"

from typing import List, TYPE_CHECKING
if TYPE_CHECKING:
    from elkpy.sushicontroller import SushiController
from .events import SushiCommandResponse
import grpc
from . import sushierrors
from . import grpc_gen

##################################
#   Sushi OSC Controller class   #
##################################


class OscController:
    """
    Class to manage OSC communication features in Sushi. It can change the OSC port and enable/disable
    OSC output for particular parameters
    """
    def __init__(self,
                 parent: 'SushiController',
                 address: str = 'localhost:51051',
                 sushi_proto_def='/usr/share/sushi/sushi_rpc.proto',
                 sushi_proto = None, sushi_grpc = None) -> None:
        """
        The constructor for the MidiController class setting up the gRPC connection with sushi.

        Parameters:
            address (str): IP address to Sushi in the uri form : 'ip-addr:port'
            sushi_proto_def (str): path to the .proto file with SUSHI gRPC services definitions
        """
        try:
            channel = grpc.insecure_channel(address)
        except AttributeError as e:
            raise TypeError(f"Parameter address = {address}. Should be a string containing the ip-address and port "
                            f"to Sushi") from e

        if sushi_proto and sushi_grpc:
            self._sushi_proto = sushi_proto
            self._sushi_grpc = sushi_grpc
        else:
            self._sushi_proto, self._sushi_grpc = grpc_gen.modules_from_proto(sushi_proto_def)
        self._stub = self._sushi_grpc.OscControllerStub(channel)
        self._parent = parent

    def get_send_port(self) -> int:
        """
        The the current port that OSC is transmitting to

        Returns:
            int: The port number
        """
        try:
            response = self._stub.GetSendPort(self._sushi_proto.GenericVoidValue())
            return response.value
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def get_receive_port(self) -> int:
        """
        The the current port that OSC is receiving on

        Returns:
            int: The port number
        """
        try:
            response = self._stub.GetReceivePort(self._sushi_proto.GenericVoidValue())
            return response.value
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def get_enabled_parameter_outputs(self) -> List[str]:
        """
        Get which parameters have OSC output enabled

        Returns:
            List[str]: List of names of the parameters for which OSC is enabled
        """
        try:
            response = self._stub.GetEnabledParameterOutputs(self._sushi_proto.GenericVoidValue())
            return [path for path in response.path]
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def enable_output_for_parameter(self, processor_id: int, parameter_id: int) -> SushiCommandResponse:
        """
        Enable OSC for a parameter

        Parameters:
            processor_id (int): The id of the processor the parameter belongs to
            parameter_id (int): The id of the parameter to enable OSC for

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.EnableOutputForParameter(self._sushi_proto.ParameterIdentifier(processor_id=processor_id, parameter_id=parameter_id))
            return self._parent.validate_command_response(cr, f"Enable output for parameter {parameter_id}")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def disable_output_for_parameter(self, processor_id: int, parameter_id: int) -> SushiCommandResponse:
        """
        Disable OSC for a parameter

        Parameters:
            processor_id (int): The id of the processor the parameter belongs to
            parameter_id (int): The id of the parameter to disble OSC for

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.DisableOutputForParameter(self._sushi_proto.ParameterIdentifier(processor_id=processor_id, parameter_id=parameter_id))
            return self._parent.validate_command_response(cr, f"Disable output for parameter {parameter_id}")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def enable_all_output(self) -> SushiCommandResponse:
        """
        Enable OSC output for all parameters

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.EnableAllOutput(self._sushi_proto.GenericVoidValue())
            return self._parent.validate_command_response(cr, "Enable all outputs")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def disable_all_output(self) -> SushiCommandResponse:
        """
        Disable OSC output for all parameters

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.DisableAllOutput(self._sushi_proto.GenericVoidValue())
            return self._parent.validate_command_response(cr, "Disable all outputs")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)



__author__ = "Maxime Gendebien"
__copyright__ = """

    Copyright 2017-2019 Modern Ancient Instruments Networked AB, dba Elk

    elkpy is free software: you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation, either version 3
    of the License, or (at your option) any later version.

    elkpy is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with elkpy.  If
    not, see <http://www.gnu.org/licenses/>.
"""
__license__ = "GPL-3.0"


import grpc.experimental.aio
import asyncio
from threading import Thread
from . import sushierrors
from . import grpc_gen
from .sushi_info_types import CommandResponseStatus

from typing import TYPE_CHECKING, Callable

if TYPE_CHECKING:
    from .sushicontroller import SushiController 


###########################################
#   Sushi Notification Controller class   #
###########################################


class NotificationController:
    """
    Class to manage subscriptions to Sushi notifications (changes, updates, ...)
    It allows the User, through simple API calls, to subscribe to any notification stream implemented in Sushi,
    and to attach call-back functions to each subscribed stream.

    (See the API section at the bottom of this class.)

    Attributes:
        address: gRPC server IP (str: ip:port)
        loop: an asynchronous event loop

    Notes:
        close() should ALWAYS be called as part of an application housekeeping/cleanup-before-shutdown routine as it
        ensure proper releasing of resources and clean joining of concurrent threads.
    """

    def __init__(
        self,
        parent,
        address="localhost:51051",
        sushi_proto_def="/usr/share/sushi/sushi_rpc.proto",
        sushi_proto=None,
        sushi_grpc=None,
        auto_update_async_command_responses=True,
    ):
        """
        The constructor for the NotificationController class setting up the gRPC connection with sushi.

        Parameters:
            address (str): 'ip-address:port' The ip-address and port at which to connect to sushi.
            sushi_proto_def (str): path to .proto file with SUSHI's gRPC services definition.
        """
        self._parent: "SushiController" = parent
        self.address = address
        if sushi_proto and sushi_grpc:
            self._sushi_proto = sushi_proto
            self._sushi_grpc = sushi_grpc
        else:
            self._sushi_proto, self._sushi_grpc = grpc_gen.modules_from_proto(
                sushi_proto_def
            )
        self.tasks = []

        if auto_update_async_command_responses:
            try:
                self.loop = asyncio.get_running_loop()
                self._async = True
                self.tasks.append(asyncio.create_task(self.update_command_responses()))
            except RuntimeError:
                self._async = False
                self.loop = asyncio.new_event_loop()
                self.notification_thread = Thread(
                    target=self._run_notification_loop, args=(self.loop,)
                )
                self.notification_thread.daemon = True
                self.notification_thread.start()
                self.tasks.append(
                    asyncio.run_coroutine_threadsafe(
                        self.update_command_responses(), self.loop
                    )
                )

    @staticmethod
    def _run_notification_loop(loop):
        """Attaches the asyncio event loop to the thread and start looping over it.
        Should not be called by the User.
        """
        asyncio.set_event_loop(loop)
        loop.run_forever()

    def close(self):
        for t in self.tasks:
            try:
                t.cancel()
            except Exception as e:
                print(e)

        if self._async:
            return
        else:
            self.loop.call_soon_threadsafe(self.loop.stop)

    def __del__(self):
        self.close()

    #################################################
    # Notification stream processing                #
    # Should not be called directly by the user.    #
    #################################################

    async def process_transport_change_notifications(self, call_back=None):
        try:
            async with grpc.experimental.aio.insecure_channel(self.address) as channel:
                stub = self._sushi_grpc.NotificationControllerStub(channel)
                stream = stub.SubscribeToTransportChanges(
                    self._sushi_proto.GenericVoidValue()
                )
                async for notification in stream:
                    # User logic here
                    if call_back and callable(call_back):
                        if asyncio.iscoroutinefunction(call_back):
                            await call_back(notification)
                        else:
                            call_back(notification)
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)
        except AttributeError:
            raise TypeError(
                f"Parameter address = {self.address}. "
                f"Should be a string containing the IP address and port to Sushi"
            )

    async def process_timing_update_notifications(self, call_back=None):
        try:
            async with grpc.experimental.aio.insecure_channel(self.address) as channel:
                stub = self._sushi_grpc.NotificationControllerStub(channel)
                stream = stub.SubscribeToEngineCpuTimingUpdates(
                    self._sushi_proto.GenericVoidValue()
                )
                async for notification in stream:
                    # User logic here
                    if call_back and callable(call_back):
                        if asyncio.iscoroutinefunction(call_back):
                            await call_back(notification)
                        else:
                            call_back(notification)
                    else:
                        raise TypeError(
                            "No valid call-back function has been provided for Timing Update "
                            "notification processing "
                        )
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)
        except AttributeError:
            raise TypeError(
                f"Parameter address = {self.address}. "
                f"Should be a string containing the IP address and port to Sushi"
            )

    async def process_track_change_notifications(self, call_back=None):
        try:
            async with grpc.experimental.aio.insecure_channel(self.address) as channel:
                stub = self._sushi_grpc.NotificationControllerStub(channel)
                stream = stub.SubscribeToTrackChanges(
                    self._sushi_proto.GenericVoidValue()
                )
                async for notification in stream:
                    # User logic here
                    if call_back and callable(call_back):
                        if asyncio.iscoroutinefunction(call_back):
                            await call_back(notification)
                        else:
                            call_back(notification)
                    else:
                        raise TypeError(
                            "No valid call-back function has been provided for Track Change "
                            "notification processing "
                        )
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)
        except AttributeError:
            raise TypeError(
                f"Parameter address = {self.address}. "
                f"Should be a string containing the IP address and port to Sushi"
            )

    async def process_processor_change_notifications(self, call_back=None):
        try:
            async with grpc.experimental.aio.insecure_channel(self.address) as channel:
                stub = self._sushi_grpc.NotificationControllerStub(channel)
                stream = stub.SubscribeToProcessorChanges(
                    self._sushi_proto.GenericVoidValue()
                )
                async for notification in stream:
                    # User logic here
                    if call_back and callable(call_back):
                        if asyncio.iscoroutinefunction(call_back):
                            await call_back(notification)
                        else:
                            call_back(notification)
                    else:
                        raise TypeError(
                            "No valid call-back function has been provided for Processor Change "
                            "notification processing "
                        )
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)
        except AttributeError:
            raise TypeError(
                f"Parameter address = {self.address}. "
                f"Should be a string containing the IP address and port to Sushi"
            )

    async def process_parameter_update_notifications(
        self, call_back=None, param_list=None
    ):
        if not param_list:
            block_list = self._sushi_proto.GenericVoidValue()
        else:
            p_list = []
            for p in param_list:
                param = self._sushi_proto.ParameterIdentifier(
                    processor_id=p[0], parameter_id=p[1]
                )
                p_list.append(param)
            block_list = self._sushi_proto.ParameterNotificationBlocklist(
                parameters=p_list
            )
        try:
            async with grpc.experimental.aio.insecure_channel(self.address) as channel:
                stub = self._sushi_grpc.NotificationControllerStub(channel)
                stream = stub.SubscribeToParameterUpdates(block_list)
                async for notification in stream:
                    # User logic here
                    if call_back and callable(call_back):
                        if asyncio.iscoroutinefunction(call_back):
                            await call_back(notification)
                        else:
                            call_back(notification)
                    else:
                        raise TypeError(
                            "No valid call-back function has been provided for Parameter Update "
                            "notification processing "
                        )
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)
        except AttributeError:
            raise TypeError(
                f"Parameter address = {self.address}. "
                f"Should be a string containing the IP address and port to Sushi"
            )

    async def process_property_update_notifications(
        self, call_back=None, property_list=None
    ):
        if not property_list:
            block_list = self._sushi_proto.GenericVoidValue()
        else:
            p_list = []
            for p in property_list:
                prop = self._sushi_proto.PropertyIdentifier(
                    processor_id=p[0], property_id=p[1]
                )
                p_list.append(prop)
            block_list = self._sushi_proto.PropertyNotificationBlocklist(
                properties=p_list
            )
        try:
            async with grpc.experimental.aio.insecure_channel(self.address) as channel:
                stub = self._sushi_grpc.NotificationControllerStub(channel)
                stream = stub.SubscribeToPropertyUpdates(block_list)
                async for notification in stream:
                    # User logic here
                    if call_back and callable(call_back):
                        if asyncio.iscoroutinefunction(call_back):
                            await call_back(notification)
                        else:
                            call_back(notification)
                    else:
                        raise TypeError(
                            "No valid call-back function has been provided for Property Change "
                            "notification processing "
                        )
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)
        except AttributeError:
            raise TypeError(
                f"Parameter address = {self.address}. "
                f"Should be a string containing the IP address and port to Sushi"
            )

    async def process_async_command_response_updates(self, call_back=None):
        try:
            async with grpc.experimental.aio.insecure_channel(self.address) as channel:
                stub = self._sushi_grpc.NotificationControllerStub(channel)
                stream = stub.SubscribeToAsyncCommandUpdates()
                async for notification in stream:
                    # User logic here
                    if call_back and callable(call_back):
                        if asyncio.iscoroutinefunction(call_back):
                            await call_back(notification)
                        else:
                            call_back(notification)
                    else:
                        raise TypeError(
                            "No valid call-back function has been provided for Async Command Response updates. "
                        )
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)
        except AttributeError:
            raise TypeError(
                f"Parameter address = {self.address}. "
                f"Should be a string containing the IP address and port to Sushi"
            )

    ####################################################
    # API : Subscription to Sushi notification streams #
    ####################################################

    def subscribe_to_transport_changes(self, cb) -> None:
        """
        Subscribes to Transport changes notification stream from Sushi
        User needs to implement their own stream consumer logic and pass it as cb.

        Parameters:
            cb: a callable that will be called for each notification received from the stream.
        """
        if self._async:
            self.tasks.append(
                asyncio.create_task(self.process_transport_change_notifications(cb))
            )
        else:
            asyncio.run_coroutine_threadsafe(
                self.process_transport_change_notifications(cb), self.loop
            )

    def subscribe_to_timing_updates(self, cb):
        """
        Subscribes to Timing update notification stream from Sushi
        User needs to implement their own stream consumer logic and pass it as cb.

        Parameters:
            cb: a callable that will be called for each notification received from the stream."""
        if self._async:
            self.tasks.append(
                asyncio.create_task(self.process_timing_update_notifications(cb))
            )
        else:
            asyncio.run_coroutine_threadsafe(
                self.process_timing_update_notifications(cb), self.loop
            )

    def subscribe_to_track_changes(self, cb):
        """
        Subscribes to Track change notification stream from Sushi.
        User needs to implement their own stream consumer logic and pass it as cb.

        Parameters:
            cb: a callable that will be called for each notification received from the stream.
        """
        if self._async:
            self.tasks.append(
                asyncio.create_task(self.process_track_change_notifications(cb))
            )
        else:
            asyncio.run_coroutine_threadsafe(
                self.process_track_change_notifications(cb), self.loop
            )

    def subscribe_to_processor_changes(self, cb):
        """
        Subscribes to Processor change notification stream from Sushi.
        User needs to implement their own stream consumer logic and pass it as cb.

        Parameters:
            cb: a callable that will be called for each notification received from the stream.
        """
        if self._async:
            self.tasks.append(
                asyncio.create_task(self.process_processor_change_notifications(cb))
            )
        else:
            asyncio.run_coroutine_threadsafe(
                self.process_processor_change_notifications(cb), self.loop
            )

    def subscribe_to_parameter_updates(self, cb, param_blocklist=None):
        """
        Subscribes to Parameter update notification stream from Sushi
        User needs to implement their own logic to process these notification in the placeholder methods below

        Parameters:
            cb: a callable that will be called for each notification received from the stream.
            param_blocklist: a list of parameter identifiers for which to block update notifications. \
                        A parameter identifier is itself a list of [processor_id: int, parameter_id: int] \
                        If no param_blocklist is passed, all parameter notifications will be subscribed to. \

        Notes to write useful callbacks:
            Notification objects have 2 attributes: parameter and value;
            Parameter itself has 2 attributes: processor_id and _parameter_id;
            ex: notification.parameter.parameter_id (gets the parameter ID)
            ex: notification.parameter.processor_id (gets the processor ID)
            ex: notification.normalized_value (gets the value normalized between 0 and 1)
            ex: notification.domain_value (gets the domain value)
            ex: notification.formatted_value (gets the value formatted as a string)
        """
        if self._async:
            self.tasks.append(
                asyncio.create_task(self.process_parameter_update_notifications(cb))
            )
        else:
            asyncio.run_coroutine_threadsafe(
                self.process_parameter_update_notifications(cb, param_blocklist),
                self.loop,
            )

    def subscribe_to_property_updates(self, cb, property_blocklist=None):
        """
        Subscribes to Property update notification stream from Sushi
        User needs to implement their own logic to process these notification in the placeholder methods below

        Parameters:
            cb: a callable that will be called for each notification received from the stream.

        Notes to write useful callbacks:
            Notification objects have 2 attributes: property and value;
            Property itself has 2 attributes: processor_id and property_id;
            ex: notification.parameter.property_id (gets the property ID)
            ex: notification.parameter.processor_id (gets the processor ID)
            ex: notification.value (gets the value)
        """
        if self._async:
            self.tasks.append(
                asyncio.create_task(
                    self.process_property_update_notifications(cb, property_blocklist)
                )
            )
        else:
            asyncio.run_coroutine_threadsafe(
                self.process_property_update_notifications(cb, property_blocklist),
                self.loop,
            )

    def subscribe_to_async_command_response(self, cb: Callable) -> None:
        """
        Subscribes to asynchronous updates to command responses.
        User may implement their own logic to process these notifications but when it comes to leveraging
        this mechanism to wait for a command completion, elkpy includes ElkpyEvents, specifically
        designed for that purpose.

        Parameters:
            cb: a callable that will be called for each notification received from the stream.
        """
        if self._async:
            self.tasks.append(
                asyncio.create_task(self.process_async_command_response_updates(cb))
            )
        else:
            asyncio.run_coroutine_threadsafe(
                self.process_async_command_response_updates(cb),
                self.loop,
            )

    #################################################
    # Internal event<->notificaton matching methods #
    #################################################
    async def update_command_responses(self) -> None:
        """Matches incoming notifications for command response updates with ElkpyEvents"""
        try:
            async with grpc.experimental.aio.insecure_channel(self.address) as channel:
                stub = self._sushi_grpc.NotificationControllerStub(channel)
                stream = stub.SubscribeToAsyncCommandUpdates(
                    self._sushi_proto.GenericVoidValue()
                )
                async for notification in stream:
                    # match with events and set them.
                    if ev := self._parent.command_responses.get(
                        notification.request_id
                    ):
                        match notification.status.status:
                            case CommandResponseStatus.SUCCESS:
                                ev.set()
                            case CommandResponseStatus.ERROR:
                                ev.error = True
                                ev.set()
                            case _:
                                print(notification.status.status)

                        self._parent.command_responses.pop(notification.request_id)
        except Exception:
            pass



__author__ = "Maxime Gendebien"
__copyright__ = """

    Copyright 2017-2019 Modern Ancient Instruments Networked AB, dba Elk

    elkpy is free software: you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation, either version 3
    of the License, or (at your option) any later version.

    elkpy is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with elkpy.  If
    not, see <http://www.gnu.org/licenses/>.
"""
__license__ = "GPL-3.0"

from .events import SushiCommandResponse
import grpc
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from elkpy.sushicontroller import SushiController

from . import sushierrors
from . import sushi_info_types as info_types
from . import grpc_gen
from typing import List

############################################
#   Sushi Audio Routing Controller class   #
############################################


class AudioRoutingController(object):
    """
    A class to control audio routing in Sushi via gRPC. It manages audio input and output
    connections between different tracks in sushi and how they connect to external inputs and outputs.

    Attributes:
        _stub (AudioRoutingControllerStub): connection stub to the gRPC audio routing interface in sushi
    """
    def __init__(self,
                 parent: 'SushiController',
                 address: str='localhost:51051',
                 sushi_proto_def: str='/usr/share/sushi/sushi_rpc.proto',
                 sushi_proto = None, sushi_grpc = None) -> None:
        """
        The constructor for the AudioRoutingController class setting up the gRPC connection with sushi.

        Parameters:
            address (str): IP address to Sushi in the uri form : 'ip-addr:port'
            sushi_proto_def (str): path to the .proto file with SUSHI gRPC services definitions
        """
        try:
            channel = grpc.insecure_channel(address)
        except AttributeError as e:
            raise TypeError(f"Parameter address = {address}. Should be a string containing the ip-address and port "
                            f"to Sushi") from e
        
        if sushi_proto and sushi_grpc:
            self._sushi_proto = sushi_proto
            self._sushi_grpc = sushi_grpc
        else:
            self._sushi_proto, self._sushi_grpc = grpc_gen.modules_from_proto(sushi_proto_def)
        self._stub = self._sushi_grpc.AudioRoutingControllerStub(channel)
        
        self._parent = parent

    def get_all_input_connections(self) -> List[info_types.AudioConnection]:
        """
        Gets a list of all input connections.

        Returns:
            List[info_types.AudioConnection]: a list of AudioConnection objects.
        """
        try:
            response = self._stub.GetAllInputConnections(self._sushi_proto.GenericVoidValue())
            return [info_types.AudioConnection(connection) for connection in response.connections]
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def get_all_output_connections(self) -> List[info_types.AudioConnection]:
        """
        Gets a list of all output connections.

        Returns:
            List[info_types.AudioConnection]: a list of AudioConnection objects.
        """
        try:
            response = self._stub.GetAllOutputConnections(self._sushi_proto.GenericVoidValue())
            return [info_types.AudioConnection(connection) for connection in response.connections]
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def get_input_connections_for_track(self, track_id: int) -> List[info_types.AudioConnection]:
        """
        Gets a list of input connections for a specific track.

        Parameters:
            track_id (int): The id of the track to get the input connections from

        Returns:
            List[info_types.AudioConnection]: a list of AudioConnection objects.
        """
        try:
            cr = self._stub.GetInputConnectionsForTrack(self._sushi_proto.TrackIdentifier(id=track_id))
            response = self._parent.validate_command_response(cr, f"Get input connections for track {track_id}")
            return [info_types.AudioConnection(connection) for connection in response.connections]
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def get_output_connections_for_track(self, track_id: int) -> List[info_types.AudioConnection]:
        """
        Gets a list of output connections for a specific track.

        Parameters:
            track_id (int): The id of the track to get the output connections from

        Returns:
            List[info_types.AudioConnection]: a list of AudioConnection objects.

        """
        try:
            cr = self._stub.GetOutputConnectionsForTrack(self._sushi_proto.TrackIdentifier(id=track_id))
            response = self._parent.validate_command_response(cr, f"Get output connections for track {track_id}")
            return [info_types.AudioConnection(connection) for connection in response.connections]
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def connect_input_channel_to_track(self, track: int, track_channel: int, engine_channel: int) -> SushiCommandResponse:
        """
        Connects an input channel to a track

        Parameters:
            track (int): The index of the track to connect to
            track_channel (int): The index of the channel on the track to connect
            engine_channel (int): The index of the channel on the engine to connect

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.ConnectInputChannelToTrack(self._sushi_proto.AudioConnection(track=self._sushi_proto.TrackIdentifier(id=track),
                                                                                    track_channel=track_channel,
                                                                                    engine_channel=engine_channel
                                                                                    ))
            return self._parent.validate_command_response(cr, f"Connect input channel to track {track}")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, f"With track id: {track}, track_channel: {track_channel} "
                                               f"and engine_channel: {engine_channel}")

    def connect_output_channel_from_track(self, track: int, track_channel: int, engine_channel: int) -> SushiCommandResponse:
        """
        Connects an output channel from a track

        Parameters:
            track (int): The index of the track to connect to
            track_channel (int): The index of the channel on the track to connect
            engine_channel (int): The index of the channel on the engine to connect

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.ConnectOutputChannelFromTrack(
                self._sushi_proto.AudioConnection(track=self._sushi_proto.TrackIdentifier(id=track),
                                                  track_channel=track_channel,
                                                  engine_channel=engine_channel
                                                  ))
            return self._parent.validate_command_response(cr, f"Connect output channel from track {track}")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, f"With track id: {track}, track_channel: {track_channel} "
                                               f"and engine_channel: {engine_channel}")

    def disconnect_input(self, track: int, track_channel: int, engine_channel: int) -> SushiCommandResponse:
        """
        Disconnects an input from a track

        Parameters:
            track (int): The index of the track to disconnect to
            track_channel (int): The index of the channel on the track to disconnect
            engine_channel (int): The index of the channel on the engine to disconnect

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.DisconnectInput(
                self._sushi_proto.AudioConnection(track=self._sushi_proto.TrackIdentifier(id=track),
                                                  track_channel=track_channel,
                                                  engine_channel=engine_channel
                                                  ))
            return self._parent.validate_command_response(cr, f"Disconnect input from track {track}")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, f"With track id: {track}, track_channel: {track_channel} "
                                               f"and engine_channel: {engine_channel}")

    def disconnect_output(self, track: int, track_channel: int, engine_channel: int) -> SushiCommandResponse:
        """
        Disconnects an output from a track

        Parameters:
            track (int): The index of the track to disconnect to
            track_channel (int): The index of the channel on the track to disconnect
            engine_channel (int): The index of the channel on the engine to disconnect

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.DisconnectOutput(
                self._sushi_proto.AudioConnection(track=self._sushi_proto.TrackIdentifier(id=track),
                                                  track_channel=track_channel,
                                                  engine_channel=engine_channel
                                                  ))
            return self._parent.validate_command_response(cr, f"Disconnect output from track {track}")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, f"With track id: {track}, track_channel: {track_channel} "
                                               f"and engine_channel: {engine_channel}")

    def disconnect_all_inputs_from_track(self, track_id: int) -> SushiCommandResponse:
        """
        Disconnects all inputs from a track

        Parameters:
            track_id (int): a track ID for which all inputs will be disconnected

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
         """
        try:
            cr = self._stub.DisconnectAllInputsFromTrack(self._sushi_proto.TrackIdentifier(id=track_id))
            return self._parent.validate_command_response(cr, f"Disconnect all inputs from track {track_id}")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, f"With track id: {track_id}")

    def disconnect_all_outputs_from_track(self, track_id: int) -> SushiCommandResponse:
        """
        Disconnects all outputs from a track

        Parameters:
            track_id (int): a track ID for which all inputs will be disconnected

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
          """
        try:
            cr = self._stub.DisconnectAllOutputsFromTrack(self._sushi_proto.TrackIdentifier(id=track_id))
            return self._parent.validate_command_response(cr, f"Disconnect all outputs from track {track_id}")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, f"With track id: {track_id}")



__author__ = "Ruben Svensson"
__copyright__ = """

    Copyright 2017-2019 Modern Ancient Instruments Networked AB, dba Elk

    elkpy is free software: you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation, either version 3
    of the License, or (at your option) any later version.

    elkpy is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with elkpy.  If
    not, see <http://www.gnu.org/licenses/>.
"""
__license__ = "GPL-3.0"

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from elkpy.sushicontroller import SushiController

import grpc

from . import sushierrors
from . import grpc_gen
from . import sushi_info_types as info_types
from .events import SushiCommandResponse

from typing import List

####################################
# Sushi audio graph controller class #
####################################


class AudioGraphController:
    """
    A class to control the audio graph in sushi via gRPC. It has functions to create, move, delete and get info
    about audio graph object like tracks and processors.

    Attributes:
        _stub (AudioGraphControllerStub): Connection stubs to the gRPC audio graph interface implemented in sushi.
    """

    def __init__(
        self,
        parent,
        address="localhost:51051",
        sushi_proto_def="/usr/share/sushi/sushi_rpc.proto",
        sushi_proto=None,
        sushi_grpc=None,
    ):
        """
        The constructor for the AudioGraphController class setting up the gRPC connection with sushi.

        Parameters:
            address (str): 'ip-addres:port' The ip-addres and port at which to connect to sushi.
            sushi_proto_def (str): path to .proto file with SUSHI's gRPC services definition
        """
        self._parent: "SushiController" = parent

        try:
            channel = grpc.insecure_channel(address)
        except AttributeError as e:
            raise TypeError(
                "Parameter address = {}. Should be a string containing the ip-address and port of sushi ('ip-address:port')".format(
                    address
                )
            ) from e

        if sushi_proto and sushi_grpc:
            self._sushi_proto = sushi_proto
            self._sushi_grpc = sushi_grpc
        else:
            self._sushi_proto, self._sushi_grpc = grpc_gen.modules_from_proto(
                sushi_proto_def
            )

        self._stub = self._sushi_grpc.AudioGraphControllerStub(channel)

    def get_all_processors(self) -> List[info_types.ProcessorInfo]:
        """
        Gets a list of all available processors.

        Returns:
            List[info_types.ProcessorInfo]: A list with the info of all the available processors.
        """
        try:
            response = self._stub.GetAllProcessors(self._sushi_proto.GenericVoidValue())
            processor_info_list = []
            for processor_info in response.processors:
                processor_info_list.append(info_types.ProcessorInfo(processor_info))
            return processor_info_list

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def get_all_tracks(self) -> List[info_types.TrackInfo]:
        """
        Gets a list of all available tracks.

        Returns:
            List[info_types.TrackInfo]: A list with the info of all the available tracks.
        """
        try:
            response = self._stub.GetAllTracks(self._sushi_proto.GenericVoidValue())

            track_info_list = []
            for track_info in response.tracks:
                track_info_list.append(info_types.TrackInfo(track_info))
            return track_info_list

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def get_track_id(self, track_name: str) -> int:
        """
        Get the id of a track from its name.

        Parameters:
            track_name (str): The name of the track.

        Returns:
            id (int): The track id in Sushi
        """
        try:
            cr = self._stub.GetTrackId(
                self._sushi_proto.GenericStringValue(value=track_name)
            )
            response = self._parent.validate_command_response(
                cr, context_info=f"Getting id for track {track_name}"
            )
            return response.id

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, "With track name: {}".format(track_name))

    def get_track_info(self, track_identifier: int) -> info_types.TrackInfo:
        """
        Get the info of a track from its id.

        Parameters:
            track_identifier (int): The id of the track to get the info from.

        Returns:
            info_types.TrackInfo: The info of the track matching the id.
        """
        try:
            cr = self._stub.GetTrackInfo(
                self._sushi_proto.TrackIdentifier(id=track_identifier)
            )
            response = self._parent.validate_command_response(
                cr, context_info=f"Getting track info for track {track_identifier}"
            )
            return info_types.TrackInfo(response.info)

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(
                e, "With track id: {}".format(track_identifier)
            )

    def get_track_processors(
        self, track_identifier: int
    ) -> List[info_types.ProcessorInfo]:
        """
        Get a list of processors assigned on the specified track.

        Parameters:
            track_identifier (int): The id of the track to get the processor list from.

        Returns:
            List[info_types.ProcessorInfo]: A list of the info of the processors assigned to the track matching the id.
        """
        try:
            cr = self._stub.GetTrackProcessors(
                self._sushi_proto.TrackIdentifier(id=track_identifier)
            )
            response = self._parent.validate_command_response(
                cr, f"Getting track processors for track {track_identifier}"
            )
            processor_info_list = []
            for processor_info in response.processors:
                processor_info_list.append(info_types.ProcessorInfo(processor_info))

            return processor_info_list

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(
                e, "With track id: {}".format(track_identifier)
            )

    def get_processor_id(self, processor_name: str) -> int:
        """
        Get the id of a processor from its name.

        Parameters:
            processor_name (str): The name of the processor to get the id from.

        Returns:
            int: The id of the processor matching the name.
        """
        try:
            cr = self._stub.GetProcessorId(
                self._sushi_proto.GenericStringValue(value=processor_name)
            )
            response = self._parent.validate_command_response(
                cr, f"Getting id for processor {processor_name}"
            )
            return response.id

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(
                e, "With processor name: {}".format(processor_name)
            )

    def get_processor_info(self, processor_identifier: int) -> info_types.ProcessorInfo:
        """
        Get the info of a processor from its id.

        Parameters:
            track_identifier (int): The id of the processor to get the info from.

        Returns:
            info_types.ProcessorInfo: The info of the processor matching the id.
        """
        try:
            cr = self._stub.GetProcessorInfo(
                self._sushi_proto.ProcessorIdentifier(id=processor_identifier)
            )
            response = self._parent.validate_command_response(
                cr, f"Getting processor info for processor {processor_identifier}"
            )
            return info_types.ProcessorInfo(response.processor)

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(
                e, "With processor id: {}".format(processor_identifier)
            )

    def get_processor_bypass_state(self, processor_identifier: int) -> bool:
        """
        Get the bypass state of the specified processor.

        Parameters:
            processor_identifier (int): The id of processor to get the bypass state from.

        Returns:
            bool: The bypass state of the processor matching the id.
        """
        try:
            cr = self._stub.GetProcessorBypassState(
                self._sushi_proto.ProcessorIdentifier(id=processor_identifier)
            )
            response = self._parent.validate_command_response(
                cr, f"Getting bypass state for processor {processor_identifier}"
            )
            return response.value

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(
                e, "With processor id: {}".format(processor_identifier)
            )

    def get_processor_state(
        self, processor_identifier: int
    ) -> info_types.ProcessorState:
        """
        Get the full state of the specified processor.

        Parameters:
            processor_identifier (int): The id of processor to get the full state from.

        Returns:
            ProcessorState: An object describing the full the processor matching the id.
        """
        try:
            cr = self._stub.GetProcessorState(
                self._sushi_proto.ProcessorIdentifier(id=processor_identifier)
            )
            response = self._parent.validate_command_response(
                cr, f"Getting the state of processor {processor_identifier}"
            )
            return response.state

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(
                e, "With processor id: {}".format(processor_identifier)
            )

    def set_processor_bypass_state(
        self, processor_identifier: int, bypass_state: bool
    ) -> SushiCommandResponse:
        """
        Set the bypass state of the specified processor.

        Parameters:
            processor_identifier (int): The id of the processor to set the bypass state of.
            bypass_sate (bool): The bypass state of the processor matching the id.

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.SetProcessorBypassState(
                self._sushi_proto.ProcessorBypassStateSetRequest(
                    processor=self._sushi_proto.ProcessorIdentifier(
                        id=processor_identifier
                    ),
                    value=bypass_state,
                )
            )
            return self._parent.validate_command_response(
                cr,
                f"Setting bypass state of processor {processor_identifier} to {bypass_state}",
            )

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(
                e,
                "With processor id: {}, bypass state: {}".format(
                    processor_identifier, bypass_state
                ),
            )

    def set_discrete_processor_state(
        self,
        processor_identifier: int,
        program_id: int | None = None,
        bypassed: bool | None = False,
        property_values: list[tuple[int, str]] = [],
        parameter_values: list[tuple[int, float]] = [],
    ) -> None:
        """
        Set the full or partial state of the specified processor.

        Parameters:
            program_id (int): The id of the program to set.
            bypassed (bool): Whether the processor should be bypassed or not.
            properties ((int, str)): A list of tuples (id, value) of string properties to set.
            parameters ((int, float)): A list of tuples (id, value) of parameter values to set.
        """
        try:
            grpc_state = self._sushi_proto.ProcessorState()

            if program_id:
                grpc_state.program_id.value = program_id
                grpc_state.program_id.has_value = True

            if bypassed is not None:
                grpc_state.bypassed.value = bypassed
                grpc_state.bypassed.has_value = True

            for property in property_values:
                grpc_property = grpc_state.properties.add()
                grpc_property.property.property_id = property[0]
                grpc_property.value = property[1]

            for parameter in parameter_values:
                grpc_parameter = grpc_state.parameters.add()
                grpc_parameter.parameter.parameter_id = parameter[0]
                grpc_parameter.value = parameter[1]

            self._stub.SetProcessorState(
                self._sushi_proto.ProcessorStateSetRequest(
                    processor=self._sushi_proto.ProcessorIdentifier(
                        id=processor_identifier
                    ),
                    state=grpc_state,
                )
            )

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(
                e, "With processor id: {}".format(processor_identifier)
            )

    def set_processor_state(
        self, processor_identifier: int, state: info_types.ProcessorState
    ) -> None:
        """
        Set the full or partial state of the specified processor from an existing state object.

        Parameters:
            state (info_types.ProcessorState): a state object either populated manually or received from a call to get_processor_state.
        """
        try:
            grpc_state = self._sushi_proto.ProcessorState()

            if state.program_id:
                grpc_state.program_id.value = state.program_id
                grpc_state.program_id.has_value = True

            if state.bypassed is not None:
                grpc_state.bypassed.value = state.bypassed
                grpc_state.bypassed.has_value = True

            for property in state.properties:
                grpc_property = grpc_state.properties.add()
                grpc_property.property.property_id = property[0]
                grpc_property.value = property[1]

            for parameter in state.parameters:
                grpc_parameter = grpc_state.parameters.add()
                grpc_parameter.parameter.parameter_id = parameter[0]
                grpc_parameter.value = parameter[1]

            if len(state.binary_data) > 0:
                grpc_state.binary_data = state.binary_data

            self._stub.SetProcessorState(
                self._sushi_proto.ProcessorStateSetRequest(
                    processor=self._sushi_proto.ProcessorIdentifier(
                        id=processor_identifier
                    ),
                    state=grpc_state,
                )
            )

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(
                e, "With processor id: {}".format(processor_identifier)
            )

    def create_track(self, name: str, channels: int) -> SushiCommandResponse:
        """
        Create a new track in sushi.

        Parameters:
            name (str): The name of the new track.
            channels (int): The number of channels to assign the new track.

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.CreateTrack(
                self._sushi_proto.CreateTrackRequest(name=name, channels=channels)
            )
            return self._parent.validate_command_response(cr, f"Creating track {name}")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(
                e, "With track name: {}, number of channels: {}".format(name, channels)
            )

    def create_multibus_track(self, name: str, buses: int) -> SushiCommandResponse:
        """
        Create a new multibus track in sushi.

        Parameters:
            name (str): The name of the new track.
            buses (int): The number of audio buses in the new track.

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.CreateMultibusTrack(
                self._sushi_proto.CreateMultibusTrackRequest(name=name, buses=buses)
            )
            return self._parent.validate_command_response(
                cr, f"Creating multibus track {name}"
            )
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(
                e, "With track name: {}, buses: {}".format(name, buses)
            )

    def create_pre_track(self, name: str) -> SushiCommandResponse:
        """
        Create a new pre track in sushi.

        Parameters:
            name (str): The name of the new track.

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.CreatePreTrack(
                self._sushi_proto.CreatePreTrackRequest(name=name)
            )
            return self._parent.validate_command_response(
                cr, f"Creating pre-track {name}"
            )
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, "With track name: {}".format(name))

    def create_post_track(self, name: str) -> SushiCommandResponse:
        """
        Create a new POST track in sushi.

        Parameters:
            name (str): The name of the new track.

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.CreatePostTrack(
                self._sushi_proto.CreatePostTrackRequest(name=name)
            )
            return self._parent.validate_command_response(
                cr, f"Creating post-track {name}"
            )
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, "With track name: {}".format(name))

    def create_processor_on_track(
        self,
        name: str,
        uid: str,
        path: str,
        processor_type: info_types.PluginType,
        track_id: int,
        before_processor: int,
        add_to_back: bool,
    ) -> SushiCommandResponse:
        """
        Create a new processor on an existing track.

        Parameters:
            name (str): The name to assign the processor. Must be unique.
            uid (str): The uid of an internal Sushi processor. Not applicable for vst2, vst3 and lv2.
            path (str): The path to the processor library (for vst2 or vst3 processors) or the URI of an installed lv2 plugin.
            processor_type (info_type.PluginType): The type of processor to create.
            track_id (int): The id of the track to add the processor to.
            before_processor (int): Which existing processor to create the new processor in front of.
            add_to_back (bool): Set to true to add the processor to the back of the processing chain on the track.

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """

        try:
            cr = self._stub.CreateProcessorOnTrack(
                self._sushi_proto.CreateProcessorRequest(
                    name=name,
                    uid=uid,
                    path=path,
                    type=self._sushi_proto.PluginType(type=processor_type),
                    track=self._sushi_proto.TrackIdentifier(id=track_id),
                    position=self._sushi_proto.ProcessorPosition(
                        add_to_back=add_to_back,
                        before_processor=self._sushi_proto.ProcessorIdentifier(
                            id=before_processor
                        ),
                    ),
                )
            )
            return self._parent.validate_command_response(
                cr, f"Creating processor {name} on track {track_id}"
            )
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(
                e,
                "With processor name: {}, uid: {}, path: {}, type: {}, id: {}, position: {}, add_to_back: {}".format(
                    name,
                    uid,
                    path,
                    processor_type,
                    track_id,
                    before_processor,
                    add_to_back,
                ),
            )

    def move_processor_on_track(
        self,
        processor: int,
        source_track: int,
        destination_track: int,
        before_processor: int,
        add_to_back: bool,
    ) -> SushiCommandResponse:
        """
        Move an existing processor.

        Parameters:
            processor (int): The id of the processor to move.
            source_track (int): The id of the track to move the processor from.
            destination_track (int): The id of the track to move the processor to.
            before_processor (int): The id of another processor to move this processor in front of.
            add_to_back (bool): Set to true to add the processor to the back of the processing chain on the track.

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.MoveProcessorOnTrack(
                self._sushi_proto.MoveProcessorRequest(
                    processor=self._sushi_proto.ProcessorIdentifier(id=processor),
                    source_track=self._sushi_proto.TrackIdentifier(id=source_track),
                    dest_track=self._sushi_proto.TrackIdentifier(id=destination_track),
                    position=self._sushi_proto.ProcessorPosition(
                        add_to_back=add_to_back,
                        before_processor=self._sushi_proto.ProcessorIdentifier(
                            id=before_processor
                        ),
                    ),
                )
            )
            return self._parent.validate_command_response(
                cr,
                f"Moving processor {processor} from track {source_track} to track {destination_track}",
            )

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(
                e,
                "With processor id: {}, source_track: {}, dest_track: {}, before_processor: {}, add_to_back: {}".format(
                    processor,
                    source_track,
                    destination_track,
                    before_processor,
                    add_to_back,
                ),
            )

    def delete_processor_from_track(
        self, processor: int, track: int
    ) -> SushiCommandResponse:
        """
        Delete an existing processor from a track.

        Parameters:
            processor (int): The id of the processor to delete.
            track (int): The id of the track that contains the processor.

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """

        try:
            cr = self._stub.DeleteProcessorFromTrack(
                self._sushi_proto.DeleteProcessorRequest(
                    processor=self._sushi_proto.ProcessorIdentifier(id=processor),
                    track=self._sushi_proto.TrackIdentifier(id=track),
                )
            )
            return self._parent.validate_command_response(
                cr, f"Deleting processor {processor} from track {track}"
            )
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(
                e, "With processor id: {}, track id: {}".format(processor, track)
            )

    def delete_track(self, track_id: int) -> SushiCommandResponse:
        """
        Delete a track.

        Parameters:
            track_id (int): The id of the track to delete.

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.DeleteTrack(self._sushi_proto.TrackIdentifier(id=track_id))
            return self._parent.validate_command_response(
                cr, f"Deleting track {track_id}"
            )
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, "With track id: {}".format(track_id))
